%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxc/alpha-lap.sl
% Title:          DEC alpha assembler
% Author:         Winfried Neun,
% Created:        January 15, 1993
% Modified:
% Status:         Experimental
%
% Copyright (c) 1993, Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load fasl-decls common))

(global '(*writingasmfile))

(fluid '(labeloffsets* currentoffset* codesize* codebase* entries*
	 forwardinternalreferences* lapreturnvalue* *writingfaslfile
	 initoffset* *pgwd *pwrds label-alist* *usefastframe))

(compiletime (ds labelp (x) (atom x)) )
% 
% Have to do someting for longwords (32 bit)
%
(compiletime (put 'store-long 'opencode
                        '((stl (reg 2) (displacement (reg 1) 0)))))
(compiletime (put 'load-long 'opencode
                        '((ldl (reg 1) (displacement (reg 1) 0)))))

(fluid '(&links-present& &function-has-quotes&))

(de lap (u)
  (prog (labeloffsets* lapreturnvalue* entries* *usefastframe)
	(unless *writingfaslfile
	  (setq currentoffset* 0)
	  (setq codebase* (gtbps 0)))

	(setq u (append u '((wordalign))))
	(setf &links-present& (or (look-for-links u)
		                  (assoc '*Id-apply u)
		                  (assoc '*call u)))

        (setq &function-has-quotes& t)
        (setq u (lapopt1 u))
        %%%%% (alpha-test-immmediatequote u)
	(setq u (pass1lap (&fillframeholes u)))
	(setq u (lapoptReformCalls u))
        (setq u (lapoptReformJumps u))

        (setq u (code-align u)) 
  
	(when *pgwd
	  (foreach x in u do (if (atom x)
		     (prin2 x)
		     (printf "          %p%n" x))))

	%
	%       findlabels calculates codesize and fixup label address
	%
	(findlabels u)

        (setq u (lapoptDEC u))
                % apply DEC's instruction scheduler

	(unless *writingfaslfile
		(unless (eq nextbps (wand nextbps (wnot 7))) (gtbps 1))
		(setq codebase* (gtbps (wquotient (wplus2 codesize!* 7) 8))))

	%
	%       generate machine code
	%
	(foreach x in u do (depositinstruction x))

	(when (not *writingfaslfile) (flushcache codebase* codesize*))

	(defineentries)

	(when (and (not *writingfaslfile) *pwrds)
	  (errorprintf "*** %p: base 16#%x, length %d bytes" 
		       (foreach x in entries* collect (first (car x))) 
		       codebase* codesize*))

	(return (if lapreturnvalue* (mkitem code-tag lapreturnvalue*) NIL))))

(de look-for-links (u)
       (cond ((null u) nil)
	     ((atom (car u)) (look-for-links (cdr u)))
	     ((eq (caar u) '*linke)
		  (cond ((flagp (caddar u) 'foreignfunction) t)
		         (t (look-for-links (cdr u)))))
	     ((eq (caar u) '*link)
		  (cond
		    ((eq  (cadar u) 'catchsetup) t) % is opencode
		         % but problems will arise in throw if no
		         % allocation takes place
		    ((memq (cadar u) '(fast-codeapply fast-idapply)) t)
		    ((get (cadar u) 'opencode) (look-for-links (cdr u)))
		    (t t)))
	     (t (look-for-links (cdr u)))
)      )

(de findlabels (u)
     (setq codesize* 0)
     (setq label-alist* nil)
     (foreach x in u do (setq codesize* (plus2 (onelaplength x) codesize*))))

(de onelaplength (x)  
   (cond ((atom x) (push (cons x (plus2 currentoffset* codesize*))
		         label-alist*) 0)
	 ((eq (car x) 'string) (times2 addressingunitsperitem 
				(strpack (size (second x))) ))
	 ((eq (car x) 'oddalign)
		 (if (eq (plus2 currentoffset* codesize*) 
			(wand (plus2 currentoffset* codesize*) (wnot 7))) 
					addressingunitsperitem 0))
	 ((eq (car x) 'wordalign)
                 (if (eq (plus2 currentoffset* codesize*)
                        (wand (plus2 currentoffset* codesize*) (wnot 7)))
                                        0 4))
	 ((eq (car x) '*entry) (setq x (second x))
		(push (cons x (plus2 currentoffset* codesize*)) label-alist*) 
		            0)
	 ((get (first x) '*op-length*) (get (first x) '*op-length*))
	 (t (error 4711 (list " No length for" x)))))

(de saveentry (x)
   (cond ((equal (second x) '***code**pointer***) 
	 (setq lapreturnvalue* 
		(if *writingfaslfile currentoffset*
		             (wplus2 codebase* currentoffset*))))
	((not *writingfaslfile)
	  (setq entries* (cons (cons (rest x) currentoffset*) entries*))
	  (unless lapreturnvalue*
	    (setq lapreturnvalue* (wplus2 codebase* currentoffset*))))
	((equal (second x) '**fasl**initcode**) 
	  (setq initoffset* currentoffset* )) 
	((flagp (second x) 'internalfunction) 
	  (put (second x) 'internalentryoffset currentoffset*))
	(t (findidnumber (second x)) 
	   (dfprintfasl (list 'putentry (mkquote (second x)) 
		         (mkquote (third x)) currentoffset*)))))

(de defineentries ()
  (foreach x in entries* do 
	   (putd (first (car x)) (second (car x)) 
		 (mkcode (wplus2 codebase* (cdr x))))))

(de labeloffset (x)
  (cond ((codep x) (if *writingfaslfile  (inf x)
                         (wdifference (inf x) codebase*)))
	((setq x (assoc x label-alist*)) (cdr x))
	(t (stderror (bldmsg "Internal entry forward reference: %r" x)))))

(de depositfloat (x)     % this will not work in cross-assembly                                   
  (progn (setq x (fltinf (second x)))
	 (depositword (floatloworder x))
	 (depositword (floathighorder x))))

(de depositbyte (x)
  (setf (byte (wplus2 codebase* currentoffset*) 0) x)
  (updatebittable 1 0)
  )

(de depositword (x)
  (putmem (wplus2 codebase* currentoffset*) x)
  (updatebittable 8 0)
  (setq currentoffset* (wplus2 currentoffset* addressingunitsperitem)))

(de deposit-relocated-word (offset)
  % Given an OFFSET from CODEBASE*, deposit a word containing the
  % absolute address of that offset.
  (putmem (wplus2 codebase* currentoffset* )
	  (wplus2 offset (if *writingfaslfile 0 codebase*)))
  (updatebittable 4 reloc-word)
  (setq currentoffset* (wplus2 currentoffset* addressingunitsperitem)))
  
(de depositwordexpression (x)  % Only limited expressions now handled                                   
  (let (y)
  (cond ((fixp x) (depositword (int2sys x)))
	((stringp x) (deposit-relocated-word 
                     (if *writingfaslfile (labeloffset x) (labeloffset x) )))
	((memq  (first x) '($fluid $global fluid global))
		        (depositvaluecelllocation (second x)))
	((equal (first x) 'mkitem) (deposititem (second x) (third x)))
	((equal (first x) 'foreignentry)
		                   (depositfunctioncelllocation (second x)))
        ((equal (first x) 'immediate)
         (cond ((and (pairp (second x))
                     (equal (first (second x)) 'mkitem))
                     (depositwordexpression (second x)))
               ((setq y (assoc (second x) label-alist*))
                (deposit-relocated-word (cdr y)))
               ((and (pairp (second x))
                     (equal (first (second x)) 'entry))
                (depositfunctioncelllocation (second (second x))))
               ((and (pairp (second x)))
                     (equal (first (second x)) 'idloc)
                (depositwordidnumber (second (second x))))

	       (t (stderror (bldmsg "Undefined label: %w" (second x))))))

	((setq y (wconstevaluable x))    (depositword (int2sys y)))
	(t  (stderror (bldmsg "Expression too complicated %r" x))))))

(de depositwordidnumber (x) 
  (cond
    ((or (not *writingfaslfile) (wleq (idinf x) 128)) (depositword (idinf X)))
    (t (putword (wplus2 codebase* currentoffset*) 0 
	       (makerelocword reloc-id-number (findidnumber x))) 
       (updatebittable 8 reloc-word))))

(de deposititem (tagpart infpart)
  (if (not *writingfaslfile)
    (depositword 
       (mkitem tagpart 
	  (cond ((labelp infpart) (wplus2 codebase* (labeloffset infpart)))
		((equal (first infpart) 'idloc)    (idinf (second infpart)))
		(t (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))))
    (progn 
	(cond ((labelp infpart) (putmem (wplus2 codebase* currentoffset*) 
		                                        % RELOC_CODE_OFFSET = 0
		                      (mkitem tagpart (labeloffset infpart))))
	      ((equal (first infpart) 'idloc) 
		 (putmem (wplus2 codebase* currentoffset*)
		         (mkitem tagpart 
		          (makerelocinf reloc-id-number
		           (findidnumber (second infpart))))))
	      (t  (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))
	(updatebittable 8 reloc-inf))
        (setq currentoffset* (iplus2 addressingunitsperitem currentoffset*))))


(de systemfaslfixup ()
  (prog (x type label place fwdir )
	(while forwardinternalreferences*
	  (setq fwdir (car forwardinternalreferences*))
	  (setq type (car fwdir) label (cadr fwdir) place (caddr fwdir))
	  (if (null (get label 'internalentryoffset))
	    (errorprintf "*** %r not defined in this module, call is incorrect"
		         (cadr fwdir))
	  (cond ((eq type 'CALL)
		(store-long (wplus2 codebase* place )
		 (wor (wshift (wshift (load-long (wplus2 codebase* place ))
				-16 ) 16)
		    (wplus2
		      (wdifference (get label 'internalentryoffset) place)
			(ashift (wshift (load-long (wplus2 codebase* place ))
				16) -16)))))
		(t  (error 0 (list "Strange forward internal reference: "
		                  fwdir)))))
	  (setq forwardinternalreferences* (cdr forwardinternalreferences*)))
	(mapobl (function (lambda (y) (remprop y 'internalentryoffset))))
))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de addressvalue (label)
  (cond (*writingfaslfile (fasladdressvalue label))
	(t (lapaddressvalue label))))

(de fasladdressvalue (label)
  (cond ((numberp label) label)
	((eq label 'nearestentry*) (difference nearestentry* currentoffset*))
	((assoc label label-alist*) 
	        (cdr (assoc label label-alist*)) )
	((atom label)  (error 0 (list "Undefined label:" label)))
	((eq (car label) 'idloc)
		(updatebittable 4 reloc-right-half)
		(findidnumber  (cadr label)))
	((eq (car label) 'xidloc) 
                (updatebittable 4 reloc-right-half)
		(wor 16#8000 (findidnumber  (cadr label))))
	((and (eq (car label) 'hi) (numberp (cadr label)))
		 (wshift (cadr label) -16))
	((and (eq (car label) 'lo) (numberp (cadr label)))
                        (wshift (wshift (cadr label) 16) -16))
        ((eqcar label 'immediate) (fasladdressvalue (cadr label)))
        ((and (eqcar label 'mkitem)
              (eqcar (caddr label) 'idloc))
                 (mkitem (cadr label) (idinf (cadr (caddr label)))))
        ((and (eqcar label 'mkitem)
              (stringp (caddr label)))
                 (mkitem (cadr label) (fasladdressvalue (caddr label))))
        ((and (eqcar label 'mkitem) (codep (caddr label))) 
              (mkitem (cadr label)  (inf (caddr label))))
	(t    (error 0 (list "Undefined label:" label)))))

(de lapaddressvalue (label)
  (cond ((numberp label) label)
	((eq label 'nearestentry*) (difference nearestentry* currentoffset*))
        ((assoc label label-alist*)
                (cdr (assoc label label-alist*)) )
        ((atom label)  (error 0 (list "Undefined label:" label)))
        ((and (eq (car label) 'hi) (numberp (cadr label)))
                        (wshift (cadr label) -16))
        ((and (eq (car label) 'lo) (numberp (cadr label)))
                        (wshift (wshift (cadr label) 16) -16))
        ((and (eq (car label) 'hi) (eqcar (cadr label) 'immediate))
                        (wshift (lapaddressvalue (cadr label)) -16))
        ((and (eq (car label) 'lo) (eqcar (cadr label) 'immediate))
                        (wand (lapaddressvalue (cadr label)) 16#ffff))
        ((eq (car label) 'hi) (wshift (wplus2 codebase*
                                        (lapaddressvalue (cadr label))) -16))
        ((eq (car label) 'lo) (wand (wplus2 codebase*
 				     (lapaddressvalue (cadr label))) 16#ffff))
        ((memq (car label) '(FLUID $FLUID GLOBAL $GLOBAL))
               (wplus2 SYMVAL (wtimes2 4 (id2int (cadr label)))))
        ((memq (car label) '(entry foreignentry))
              (wplus2 SYMFNC (wtimes2 4  (id2int (cadr label)))))
        ((eq (car label) 'idloc) (id2int (cadr label)))
        ((eq (car label) 'xidloc) (wtimes2 8 (id2int (cadr label)) ))
        ((eqcar label 'immediate) (lapaddressvalue (cadr label)))
        ((and (eqcar label 'mkitem)
              (eqcar (caddr label) 'idloc))
                 (mkitem (cadr label) (idinf (cadr (caddr label)))))
        ((and (eqcar label 'mkitem) (stringp (caddr label)))
              (mkitem (cadr label) (wplus2 codebase*
					 (lapaddressvalue (caddr label)))))
        ((and (eqcar label 'mkitem) (codep (caddr label)))
              (mkitem (cadr label)  (inf (caddr label))))
        (t    (error 0 (list "Undefined label:" label)))))
 

(de jumpaddressvalue (label)
  (cond (*writingfaslfile (fasljumpaddressvalue label))
	(t (lapjumpaddressvalue label))))

(de lapjumpaddressvalue (label)
  (cond ((numberp label) label)
	((assoc label label-alist*)
              (wdifference (cdr (assoc label label-alist*)) currentoffset*))
	((atom label)  (error 0 (list "Undefined label:" label)))
	((eq (car label) 'foreignentry)
	      (getmem (wplus2 SYMFNC (wtimes2 4 (id2int (cadr label))))))
	(t    (error 0 (list "Undefined label:" label)))))

(define-constant reloc-call 2)

(de fasljumpaddressvalue (label)
  (cond ((numberp label) label)
	((assoc label label-alist*)
	       (wdifference (cdr (assoc label label-alist*)) currentoffset*))
	((atom label)  (error 0 (list "Undefined label:" label)))
	((eq (car label) 'internalentry)
	     (setq label (cadr label))  
	     (cond ((assoc label label-alist*) nearestentry*)
		   ((get label 'internalentryoffset)
		      (get label 'internalentryoffset))
		   (t (push `(call ,label ,currentoffset*)
		                forwardinternalreferences*) currentoffset*)))
	(t  (error 0 (list "Undefined label:" label)))))

(de forward-internal-reference-p (label)
  (not (or (assoc label label-alist!*)
	   (get label 'internalentryoffset))))

(de code-print (x)
  (for (from i 0 7)
       (do (printf "%x" (wand (wshift  x (wtimes2 4 (wdifference i 7))) 15)))))

%%      store-xxxx is used for saving code into BPS space. 

(de store-word (word) (store-long (plus2 codebase* currentoffset*) word))

(de store-byte (byte)     (depositbyte byte))

(de store-word-expression (word)  (depositwordexpression word))

(de store-double (word)           (depositfloat `(FLOAT ,word)))

(de store-relocatable-movex (code)   (deposit-relocatable-movex code))
