%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PSC:APOLLO-ASM.SL
% Description:  Apollo specific information for LAP-TO-ASM
% Author:       
% Created:      16-Jan-1982
% Modified:     13-Dec-84 12:59:04 (Vicki O'Day)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
% 14-Oct-87 (Kessler)
%  Put the e<<24 back, the assembler seems to prefer it.
%  Changed asmoutlap1 so it dumps extern.d on external entries.
% 10-Sep-87 (Michael Hucka)
%  Changed (put 'MkItem ...) to use %e*0x10000000 instead of %e<<24.
% 28-Oct-86 (Michael Hucka)
%  Changed references to Sun-specific files to be to Apollo-specific files.
% 20-Sep-86 (Leigh Stoller)
%  Copied from Bobcat compiler directory and modified to spit out Sun assembly
%  syntax using the 3.2 version as a guide. The jumpon function needed a jmp
%  instead of a bra so that it would not be shrunk by the assembler. Added the
%  sun-mnemonic-change-table to convert from Bobcat ops to Sun formated ops.
% 09-Aug-86 (Leigh Stoller)
%  Added two new temp A regs called t5,t6. Also modified the ordering so that
%  t0 is a1 and t6 is a5.
% 24-Jul-86 (Leigh Stoller)
%  Added print name of new Temp reg T0 to the reg deflists.
% 13-Dec-84 12:58:33 (Vicki O'Day)
%  Fixed typo in the name of the function OperandPrintIndirect.
% 21-Jun-84 (Vicki O'Day)
%  Duplicated Bob Morain's changes for HP-UX
% 26-Apr-84 08:55:22 (Brian Beach)
%  Removed duplicate definition of ASMOUT.  Updated version in LAP-TO-ASM.
% 30-Mar-84 09:31:50 (Brian Beach)
%  Added *FOREIGNLINK function from HP-CMAC, since it is used only by
%  the cross compiler.  Noted that CHIPMUNK-KLUDGE has been removed somewhere
%  along the line.  Now defined in garbage collector where it belongs.
% 07-Mar-83 09:36:42 (Brian Beach)
%  Made constant usage conform to new model.
% 15-Dec-83 13:15:49 (Tim Tillson)
%   Added header to Sam's D-register model version
% 22-Jul-83  Nancy Kendzierski
%   Changed use of "deferred" (in a comment) to "indirect".
%  1-Jul-83 Steve Lowder
%   Changed names of two procedures to reflect the changes in TLM 
%    addressing modes. PrintAutoIncrement was changed to PrintPostIncrement
%    and PrintAutoDecrement was changed to PrintPreIncrement.
%  10-Mar-83  Nancy Kendzierski
%   Added size specifications to InitFileNameFormat!*, DataFileNameFormat!*,
%    and CodeFileNameFormat!* to allow building on local disks as well as
%    an SRM.
%  25-Jan-83  Nancy Kendzierski
%    In ASMOUT, replaced EVIN with PATHIN to allow building on a "connected
%     directory" other than PHP:.
%  18-Jan-83  Gwyn Osnos
%   mods to ASMOUT for uppercase filenames
%  16-Sep-82
%   mods for OS 2.0
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load if-system))

(Fluid '(CodeFileNameFormat* 
         DataFileNameFormat* 
         InitFileNameFormat* 
         InputSymFile* 
         OutputSymFile* 
         CommentFormat* 
         GlobalDataFileName* 
         LabelFormat* 
         ExternalDeclarationFormat* 
         ExportedDeclarationFormat* 
         FullWordFormat* 
         HalfWordFormat* 
         ReserveDataBlockFormat* 
         ReserveZeroBlockFormat* 
         DefinedFunctionCellFormat* 
         UndefinedFunctionCellInstructions* 
         MainEntryPointName* 
         CodeOut* 
         DataOut* 
         *Lower 
         ASMOpenParen* 
         ASMCloseParen* 
         ModuleName* 
         NumericRegisterNames* 
         ForeignExternList* 
         DataProcState* 
         PathIn*))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This assembler produces two files: a code file and a data file. Both  %
% of these files are assembled to reside in the HP dataspace            %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%FORMATS%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq *DeclareBeforeUse t)

(setq CodeFileNameFormat* "%w.s")        % The name of the code segment file
(setq DataFileNameFormat* "d%w.s")       % postfix the "d" so that the files
                                         % [20] is a reasonable guess at aU
                                         % size specification

(setq InitFileNameFormat* "%w.init")     % [20] is a reasonable guess at a
                                         % size specification; it may need to
                                         % be changed.

(setq InputSymFile* "apollo.sym")           % default for full-kernel-build
(setq OutputSymFile* "apollo.sym")

(setq MainEntryPointName* '!m!a!i!n)  % chose a simple default
                                          % main procedure name

(setq NumericRegisterNames* '[nil "d0" "d1" "d2" "d3" "d4" "d5" "d6" "d7" 
                                "a0" "a1" "a2" "a3" "a4" "a5"]) %  scs/vo

(setq LabelFormat* "%w     equ *%n")
(setq CommentFormat* "* %p%n")          % Comments begin with an asterisk
                                        % will group alphabetically

(setq ExportedDeclarationFormat* "        entry.d %w%n")
(setq ExternalDeclarationFormat* "        extern.d %w%n") % All in DATA space

(setq FullWordFormat* "  ac.l %e%n")     % FullWord expects %e for parameter
(setq HalfWordFormat* "  dc.w %e%n")    % Will EVAL formatter

(setq ReserveDataBlockFormat* "  ds.b %e%n")
% This does *not* make zero blocks, however, the Sun manuals
% promise that a.out memory is init'ed to 0s
% Changed below to be like Vax version, so heap will be in bss. bao
%(setq ReserveZeroBlockFormat* "   .comm %w,%e%n nop%n")
(setq ReserveZeroBlockFormat* "%w  ds.b %e%n")

(put 'MkItem 'ASMExpressionFormat "((%e<<24)+%e)")   % MkItem(Tag, Inf)
%
%(put 'MkItem 'ASMExpressionFormat "((%e*0x1000000)+%e)") % MkItem(Tag, Inf)

%(setq DefinedFunctionCellFormat* " jmp %w%n")   %/ Must be LONG
%
(setq DefinedFunctionCellFormat* "  dc.w $4EF9%n   ac %w%n")

% SymFnc + 6*IdLoc(UndefinedFunction)
%(setq UndefinedFunctionCellInstructions*        
%       '((JSR (plus2 SymFnc 
%       (times2 6                        % For relative JMP
%             (IDLoc UndefinedFunction))))))
%
% Deposit directly
(setq UndefinedFunctionCellInstructions* 
  '((halfword 16#4EB9)		% Explicit JSR in ABS mode
    (fullword (plus2 SymFnc
         (times2 6		% For relative JMP
         (IDLoc UndefinedFunction)))))
)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%LISTS and CONSTANT DEFINITIONS%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq ASMOpenParen* "(")
(setq ASMCloseParen* ")")

(DefList '((LAnd &) 
	   (LOr |)
	   (LSH <<)) 'BinaryASMOp)


(DefList '((nil "a6")                     % This reg always holds NIL.     
  (t0 "a1")
  (t1 "a2") 
  (t2 "a3") 
  (t3 "d6") 
  (t4 "d7") 
  (t5 "a4")
  (t6 "a5")
          (db "a5")                       % HP convention
          (sb "a6")                       % HP convention
          (sp "a7")                       % HP convention
          (st "a7")                       % Stack Pointer
  (a0 "a0")                               % Normal names
  (a1 "a1")                               % Normal names
  (a2 "a2")                               % Normal names
  (a3 "a3")                               % Normal names
  (a4 "a4")                               % Normal names
  (a5 "a5")                               % Normal names
  (a6 "a6")                               % Normal names
  (a7 "a7")                               % Normal names
  (d0 "d0")                               % Normal names
  (d1 "d1")                               % Normal names
  (d2 "d2")                               % Normal names
  (d3 "d3")                               % Normal names
  (d4 "d4")                               % Normal names
  (d5 "d5")                               % Normal names
  (d6 "d6")                               % Normal names
  (d7 "d7"))                              % Normal names
  'RegisterName)

(DefList '((nil "a6")                     % (Reg NIL)   
  (t0 "a1")
  (t1 "a2") 
  (t2 "a3") 
  (t5 "a4")
  (t6 "a5")
          (db "a5")                       % HP convention
          (sb "a6")                       % HP convention
          (sp "a7")                       % HP convention
          (st "a7")                       % Stack Pointer
  (a0 "a0")                               % Normal names
  (a1 "a1")                               % Normal names
  (a2 "a2")                               % Normal names
  (a3 "a3")                               % Normal names
  (a4 "a4")                               % Normal names
  (a5 "a5")                               % Normal names
  (a6 "a6")                               % Normal names
  (a7 "a7"))                              % Normal names
  'ARegisterName)


(DefList 
  '((t3 "d6") 
  (t4 "d7") 
  (d0 "d0")                               % Normal names
  (d1 "d1")                               % Normal names
  (d2 "d2")                               % Normal names
  (d3 "d3")                               % Normal names
  (d4 "d4")                               % Normal names
  (d5 "d5")                               % Normal names
  (d6 "d6")                               % Normal names
  (d7 "d7")) 'DRegisterName)

(setq DataProcState* 'data)

% This patch is until init files can be read
%
% (CopyD 'OldSaveUncompiledExpression 'SaveUnCompiledExpression) %Save previous
% (CopyD 'SaveUncompiledExpression 'SaveForCompilation) %/Now clobber

% This is redefined from LAP-TO-ASM

(de ASMPseudoPrintFloat (x) 
  (progn (cond ((equal (cadr x) 0.00000E+000) 
                (printf " dc.l 0%n dc.l 0%n")) 
               ((equal (cadr x) 1.0) 
                (printf " dc.l 0x3FF00000%n dc.l 0%n")) 
               (t 
                (StdError 
                 (BldMsg "Floating point constant %r cannot be used" (cadr x)))))))

%%% This put is to associate the above code with (float x) in LAP

(put 'Float 'ASMPseudoOp 'ASMPseudoPrintFloat)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%CODE HEADERS AND TRAILERS%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de CodeFileHeader nil                  % Pure Code Segment
  (cond ((equal ModuleName* "main")
	 (CodePrintF "  program MAIN,z0001%n")
	 (CodePrintF "  data%n")
	 (setq DataProcState* 'data)
	 (CodePrintF "* this is the start of execution of the program%n")
   	 (CodePrintF "* The following is the standard start of program or routine%n")

   	 (CodeDeclareExternal 'SYMVAL)                 %/ Issue EXTERN.D early
   	 (CodeDeclareExternal 'SYMFNC)                 %/ Issue EXTERN.D early

   	 (CodePrintF "z0001    jsr   main              Call Main routine%n")
   	 (CodePrintF "* now return to OS%n")
   	 (CodePrintF "        dc.w $41F9%n")
   	 (CodePrintF "        extern.d pgm_$exit%n")
   	 (CodePrintF "        ac.l pgm_$exit%n")
   	 (CodePrintF "        jsr     (a0)%n")
   	 (CodePrintF "        rts                     return%n")
	 (setq ForeignExternList* nil))
	(t
	 (CodePrintF "   module %w,m0000%n" ModuleName*)
                                %/ Kludge, since ModuleName set in ASMOUT
	 (CodePrintF "* this is an Independent Module %n")
	 (CodePrintF "  data%n")
	 (setq DataProcState* 'data)
	 (CodeDeclareExternal 'SYMVAL)                 %/ Issue EXTERN.D early
	 (CodeDeclareExternal 'SYMFNC)                 %/ Issue EXTERN.D early
	 (setq ForeignExternList* nil))))
                
(de DataFileHeader nil 
  (DataPrintF "       module D%w%n" ModuleName*)
  (DataPrintF "       data%n"))

(de DataFileTrailer nil 
  (DataPrintF "  end%n"))

(de CodeFileTrailer nil 
  (CodePrintF "  end%n"))

%*** (mh) This isn't in the 3.2 apollo version; is it necessary at all?
(de CheckForeignExtern (Fn) 
  (cond
   ((not (Memq Fn ForeignExternList*))
    (setq ForeignExternList* (cons Fn ForeignExternList*)))))

(de CodeBlockHeader nil nil)            %/ Chuck this?
    
(de CodeBlockTrailer nil nil)
    
(de DataAlignFullWord nil nil)
  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%`%%
% PRINT PROCEDURES%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


(de PrintString (S) 
 (prog (n) 
    (cond ((neq (Size S) -1) 
     (progn (setq n (Size S)) 
       (printf " dc.b ") 
       (PrintExpression (Indx S 0)) 
       (for (from i 1 n 1) (do (PrintByte!, (Indx S i)))) 
       (PrintByte!, 0) 
       (cond ((equal (Remainder n 2) 1) (PrintByte!, 0))) 
       (Terpri) nil)) (t 
    (progn 
       (printf " dc.b ") 
       (PrintExpression 0) 
       (PrintByte!, 0) 
       (Terpri) nil)))))
 
(de PrintByte!, (x) 
  (progn (cond
	  ((greaterp (POSN) 60)
	   (printf "%n dc.b "))
	  (t (Prin2 ",")))
	 (PrintExpression x)))
    
(de TruncateString (S n) 
  (cond
   ((leq (Size S) n) S)
   (t 
    (SUB S 0 n))))

(de PrintByteList (L) 
    (foreach x in L do (PrintByte x)))

(de PrintByte (x) 
  (progn (printf " dc.b ") 
	 (PrintExpression x) 
	 (Prin2 $EOL$)))

(de PrintHalfWordList (L) 
    (foreach x in L do (PrintHalfWord x)))

(de PrintHalfWord (x) 
  (progn (printf " dc.w ") 
	 (PrintExpression x) 
	 (Prin2 $EOL$)))

(de PrintHalfWords (x) 
  (progn (RplacA PrintExpressionFormPointer* x) 
	 (printf HalfWordFormat* PrintExpressionForm*)))

(put 'HalfWord 'asmpseudoop 'ASMPseudoPrintHalfWords)

(de ASMPseudoPrintHalfWords (x) 
    (foreach Y in (cdr x) do (PrintHalfWord Y)))

(fluid '(*apollo-mnemonic-change-table*))

%%% Print out an opcode id (derived from LAP)
(de PrintOpcode (opcode)
  (prin2 opcode))

%%% Taken from PU:chars.lsp, this routine converts a character (a small
%%% number) from uppercase to lowercase.  This version is considerably
%%% simplified from the PU: one so that it works in a bare PSL (a primitive
%%% if there ever was one)

(de asm-char-downcase (c)
  (if (not (or (lessp c (char A)) (greaterp c (char Z))))
      (plus (char !a) (difference c (char !A)))
      c
  )
)

(de SpecialActionForMainEntryPoint nil 
  (progn (CodePrintLabel MainEntryPointName*) 
	 (CodeDeclareExported MainEntryPointName*)))


%%% Predicate to decide whether the given id or string is a valid
%%% one to the machine's assembler (though what lap-to-asm does if
%%% this predicate fails is beyond me!)


(de ASMSymbolP (x) 
  (ApolloSymbolP (cond ((IDP x) (ID2String x)) (t x))))

(de ApolloSymbolP (x) 
  (prog (n C i) 
	(setq n (Size x)) 
	(setq C (Indx x 0)) 
	(cond
	 ((not (or (and (geq C (char a))
			(leq C (char z)))
		   (equal C (char _))
		   (equal C (char $))	       %*** (mh) may be wrong
		   (and (geq C (char !a))
			(leq C (char !z)))))
	  (return nil)))
    (setq i 1) 
    Loop 
    (cond ((greaterp i n) (return t))) 
    (setq C (Indx x i)) 
    (cond
     ((not (or (and (geq C (char a))
		    (leq C (char z)))
	       (equal C (char _))
	       (equal C (char $))  %*** (mh) may be wrong C = Char !$
	       (and (geq (or C) (char !0))
		    (leq C (char !9))) 
	       (and (geq C (char !a))
		    (leq C (char !z)))))
      (return nil)))
    (setq i (plus i 1)) 
    (go Loop)))

(de PrintNumericOperand (x) 
   (printf "#%w" x))


(de OperandPrintIndirect (x)            % (Indirect x)
         (progn (setq x (cadr x)) 
            (Prin2 "(") 
            (PrintOperand x) 
            (Prin2 ")")))

(put 'Indirect 'OperandPrintFunction 'OperandPrintIndirect)


(de OperandPrintDisplacement (x)        % (Displacement (reg x) disp)
  (progn (setq x (cdr x)) 
	 (PrintExpression (cadr x)) 
	 (Prin2 "(")
	 (PrintOperand (car x)) 
	 (Prin2 ")")))

(put 'displacement 'OperandPrintFunction 'OperandPrintDisplacement)


(de OperandPrintIndexed (x) 
          % (Indexed (reg y) (displacement (reg x) disp))
  (let ((part2 (third x)))
       (PrintExpression (third part2))
       (prin2 "(")
       (PrintRegister (cadr part2))
       (prin2 ",")
       (PrintRegister (cadr x))
       (prin2 ".l)")))

(put 'Indexed 'OperandPrintFunction 'OperandPrintIndexed)


(de OperandPrintImmediate (x)           % (Immediate x) % for ADDRESS
  (PrintExpression (cadr x)))

(put 'Immediate 'OperandPrintFunction 'OperandPrintImmediate)


(de OperandPrintPostIncrement (x)       % (PostIncrement x)
  (progn (Prin2 "(") 
	 (PrintOperand (cadr x)) 
	 (Prin2 ")+")))

(put 'PostIncrement 'OperandPrintFunction 'OperandPrintPostIncrement)

(de OperandPrintRegList (x)             % (Reglist x)
  (progn (setq x (cdr x)) 
	 (PrintOperand (car x)) 
	 (setq x (cdr x)) 
	 (While x 
		(progn (Prin2 "/") 
		       (PrintOperand (car x)) 
		       (setq x (cdr x)))) nil))

(put 'RegList 'OperandPrintFunction 'OperandPrintRegList)

(de OperandPrintPreDecrement (x)        % (PreDecrement x)
(progn (Prin2 "-(") 
    (PrintOperand (cadr x)) 
    (Prin2 ")")))

(put 'PreDecrement 'OperandPrintFunction 'OperandPrintPreDecrement)

(de OperandPrintAbsolute (x)            % (Absolute x)
  (PrintExpression (cadr x)))

(put 'Absolute 'OperandPrintFunction 'OperandPrintAbsolute)

(de OperandPrintForeignEntry (x)        % (ForeignEntry FcnName)
    (prog (!*lower)			%*** (mh) apollo doesn't need '_'
	(setq !*lower t)
%	(Prin2 '_) 
	(Prin2 (cadr x))))
  
(put 'ForeignEntry 'OperandPrintFunction 'OperandPrintForeignEntry)

(Fluid '(ResultingCode*))

(de MCPrint (x)                         % Echo of MC's
 (CodePrintF "* %p%n" x))

(de InstructionPrint (x) 
 (CodePrintF "*%p%n" x))

(de *cerror (x) 
 (prog (i) 
    (setq i (wrs nil)) 
    (printf "%n *** CERROR: %r %n " x) 
    (wrs i) 
    (return (list (list 'cerror x)))))
 

(put 'cerror 'asmpseudoop 'printcomment)

(DefCMacro *cerror)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% A HORRID patch for BRA/BRA!.W incompatibility between
% KERNEL and LAP
% In Kernel, BRA means word sized branch
% IN LAp and FASL BRA seems to mean bytes sized (?) even though
% must be coerced somewhere
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%@%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% This version to be used ONLY in SUN-CROSS

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Since all code is moved into data space, the relative jmp used by the
% JUMPON cmacro no longer works.  It is changed here to first do an
% adda.l #l0028,a0
%jmp 2(a0)
% Note this adda.l must be in-line coded because the immediate won't get
% relocated otherwise (i.e., must be and address const)
% Ultimatly this belongs in HP9836-CMAC, but since it's in RLISP>???
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% FOR NOW, THIS ASSUMES 'REGISTER' IS A D REGISTER

(de *JumpOn (Register LowerBound UpperBound LabelList) 
  (prog (BaseLbl EndLbl Result) 
	(setq BaseLbl (Gensym)) 
	(setq EndLbl (Gensym)) 
	(setq Result (cons nil nil)) 
	(cond 
	 ((ZeroP  LowerBound)
	  (TConc Result (list 'tst!.l Register))) 
	 ((QInumP LowerBound)
	  (TConc Result (list '*WDifference Register LowerBound)))
	 (T
	  (TConc Result (list 'sub!.l LowerBound Register))))
	(TConc Result (list 'blt EndLbl)) 
	(TConc Result (list 'cmp!.l
			    (difference UpperBound LowerBound) Register)) 
	(TConc Result (list 'bgt EndLbl)) 
	(TConc Result (list 'lsl!.l 2 Register)) 
	(TConc Result (list 'HALFWORD 16#4EFB))      % JMP BaseLbl(Dn.L)
	(TConc Result (list 'HALFWORD (iplus2 16#0802 
					      (ilsh (registernumber 
						     (cadr register)) 12))))
	(TConc Result BaseLbl) 
	(foreach x in LabelList do
		 (TConc Result (list 'jmp (cadr x))))		 
	(TConc Result EndLbl) 
	(return (car Result))))

%% Changed so we print out the external declaration for any forenign entry 
%% a jsr.
(de asmoutlap1 (x)
  (prog (fn)
        (return (cond ((stringp x) (printlabel x))
                      ((atom x) (printlabel (findlocallabel x)))
                      ((setq fn (get (car x) 'asmpseudoop)) 
                       (apply fn (list x)))
                      (t
		       % instruction output form is:
      % "space" <opcode> [ "space" <operand> { "comma" <operand> } ] "newline"
                                                                           
      (cond ((and (equal (car x) 'jsr)
                  (pairp (cadr x))
                  (equal (car (cadr x)) 'foreignentry))
	     (prog (*lower)
               (setq *lower t)
	       (codedeclareexternal (cadr (cadr x))))))
      (progn (prin2 '! )
	     % Space                                    
	     (printopcode (car x))
	     (setq x (cdr x))
	     (unless (null x)
		     (prin2 '! )
		     % SPACE                                  
		     (printoperand (car x))
		     (foreach u in (cdr x) do 
			      (progn (prin2 '!,)
				     % COMMA                          
				     (printoperand u))))
	     (prin2 !$eol!$)))))))
