%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         OPENCODEDFUNCTIONS.SL
%
% Description:  opencoded functions in PSL compiler (Convex PSL3.4)
%
% Author:       W.Neun , ZIB Berlin
%
% Created:      28-Oct-1988
%
% Status:       Experimental (Do Not Distribute)
%
% Copyright (c) 1988, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% These functions where tagged as open coded in the CRAY PSL3.4
%  BYTE
%  STRBYT
%  HALFWORD
%  PUTBYTE
%  PUTSTRBYT
%  PUTHALFWORD
 
%  LONGTIMES2         %Functions using 46-bit Integer Arithmetic
%  LONGQUOTIENT
%  LONGDIVIDE
%  LONGDIVIDE-QUOTIENT
%  LONGDIVIDE-REMAINDER
 
%  Clear-Extraregs
%  catchsetup
%  !%uncatch
%  idapplye
%  idapplye2
%  (enable-operand-range-interrupt)   % XMP only
%  (disable-operand-range-interrupt)  % XMP only
%  (enable-floatingpoint-interrupt)
%  (disable-floatingpoint-interrupt)
 
(compiletime (load if-system))
 
(deflist '(
          (putbyte
                   ((*MOVE    (REG 1) (REG A1))
                    (*MOVE    (REG 2) (REG a5))
                    (*Wplus2  (REG a1) (reg a5))
                    (st!.b    (reg 3) (indexed (Reg A1) 0) )))
 
          (byte
                   ((*MOVE    (REG 1) (REG A1))
                    (*MOVE    (REG 2) (REG a5))
                    (*Wplus2  (REG a1) (reg a5))
                    (ld!.b    (indexed (Reg A1) 0) (reg 1))
                    (*wand    (reg 1) 255)))
 
          (putstrbyt
                   ((*MOVE    (REG 1) (REG A1))
                    (*MOVE    (REG 2) (REG a5))
                    (*Wplus2  (REG a1) (reg a5))
                    (st!.b    (reg 3) (indexed (Reg A1) 4) )))
 
          (strbyt
                   ((*MOVE    (REG 1) (REG A1))
                    (*MOVE    (REG 2) (REG a5))
                    (*Wplus2  (REG a1) (reg a5))
                    (ld!.b    (indexed (Reg A1) 4) (reg 1))
                    (*wand    (reg 1) 255)))
 
          (puthalfword
                   ((*MOVE    (REG 1) (REG A1))
                    (*MOVE    (REG 2) (REG a5))
                    (*wshift  (reg a5) 1)   % * 2 (byteaddress)
                    (*Wplus2  (REG a1) (reg a5))
                    (st!.h    (reg 3) (indexed (Reg A1) 0) )))
 
          (halfword
                   ((*MOVE    (REG 1) (REG A1))
                    (*MOVE    (REG 2) (REG a5))
                    (*wshift  (reg a5) 1)   % * 2 (byteaddress)
                    (*Wplus2  (REG a1) (reg a5))
                    (ld!.h    (indexed (Reg A1) 0) (reg 1))
                        % sign extension for hashtable access
                    (cvth!.w   (reg 1) (reg 1)) ))
 
          (clear-extraregs
                   ((*VsetVl  10)
                    (*Move 0 (Reg S6))
                    (*Move ($fluid Argumentblock) (Reg A1))
                    (*Vset (vreg 1)  (Reg S6))
                    (*VStore (vreg 1) (Reg A1) 1) ))
    ) 'opencode)
 
(put 'idapplye 'exitopencode
                         '((*Move (Reg 2) (Reg s7))
                           (*wshift (Reg s7) 3)       % Byte address
                           (*Move ($fluid symfnc) (reg a1))
                           (*Move (Reg s7) (reg a5))
                           (*wplus2 (Reg a1) (Reg a5))
                           (ld!w (indexed (reg a1) 2) (reg a5))
                           (*wshift (Reg s7) 4)
                           (*move 1 (reg s6))
                           (*wor  (Reg s6) (Reg s7))
                           (*Move (reg s6) (reg nargReg))
                           (jmp (Reg a5))))
 
(put 'idapplye2 'exitopencode
                         '((*Move (Reg 2) (Reg s7))
                           (*wshift (Reg s7) 3)       % Byte address
                           (*Move ($fluid symfnc) (reg a1))
                           (*Move (Reg s7) (reg a5))
                           (*wplus2 (Reg a1) (Reg a5))
                           (ld!w (indexed (reg a1) 2) (reg a5))
                           (*wshift (Reg s7) 4)
                           (*move 2 (reg s6))
                           (*wor  (Reg s6) (Reg s7))
                           (*Move (reg s6) (reg nargReg))
                           (jmp (Reg a5))))
 
%--------------------------------------------------------------
% floats

(put '*fplus2 'opencode 
                      '((*move (reg 2)(reg a1))
                        (ld!.l (indexed (reg a1) 0) (reg 4))
                        (*move (reg 3)(reg a1)) 
                        (ld!.l (indexed (reg a1) 0) (reg 2))   
                        (add!.d (reg 2)(reg 4))   
                        (*move (reg 1)(reg a1)) 
                        (st!.l (reg 4) (indexed (reg a1) 0))))

(put '*fdifference 'opencode 
                      '((*move (reg 2)(reg a1))
                        (ld!.l (indexed (reg a1) 0) (reg 4))
                        (*move (reg 3)(reg a1)) 
                        (ld!.l (indexed (reg a1) 0) (reg 2))   
                        (sub!.d (reg 2)(reg 4))   
                        (*move (reg 1)(reg a1)) 
                        (st!.l (reg 4) (indexed (reg a1) 0))))

(put '*ftimes2 'opencode 
                      '((*move (reg 2)(reg a1))
                        (ld!.l (indexed (reg a1) 0) (reg 4))
                        (*move (reg 3)(reg a1)) 
                        (ld!.l (indexed (reg a1) 0) (reg 2))   
                        (mul!.d (reg 2)(reg 4))   
                        (*move (reg 1)(reg a1)) 
                        (st!.l (reg 4) (indexed (reg a1) 0))))
 
(put '*fquotient 'opencode  
                      '((*move (reg 2)(reg a1))
                        (ld!.l (indexed (reg a1) 0) (reg 4))  
                        (*move (reg 3)(reg a1))
                        (ld!.l (indexed (reg a1) 0) (reg 2))  
                        (div!.d (reg 2)(reg 4))  
                        (*move (reg 1)(reg a1))
                        (st!.l (reg 4) (indexed (reg a1) 0)))) 
 
(put '*wfloat 'opencode 
                      '((cvtw!.s (reg 2)(reg 2))
                        (cvts!.d (reg 2)(reg 2))
                        (*move (reg 1)(reg a1))
                        (st!.l (reg 2)(indexed (reg a1) 0))))
 
(put '*wfix 'opencode
                      '((*move (reg 1)(reg a1))
                        (ld!.l (indexed (reg a1) 0) (reg 1))
                        (cvtd!.s (reg 1)(reg 1))
                        (cvts!.l (reg 1)(reg 1))))
 
(put '*fassign 'opencode
                      '((*move (reg 2)(reg a1))
                        (ld!.l (indexed (reg a1) 0) (reg 3))
                        (*move (reg 1)(reg a1))
                        (st!.l (reg 3) (indexed (reg a1) 0))))
 
(put '**fgreaterp 'opencode
                      '((*move (reg 1)(reg a1))
                        (ld!.l (indexed (reg a1) 4) (reg 1))
                        (*move (reg 2)(reg a1))
                        (ld!.l (indexed (reg a1) 4) (reg 2))
                        (lt!.d (reg 2) (reg 1))
                        (*move (fluid t) (reg 1))
                        (brs!.t  *donefgreaterp*)
                        (mov!.w (reg nil) (reg 1))
                        *donefgreaterp*))

(put '**flessp 'opencode
                      '((*move (reg 1)(reg a1))
                        (ld!.l (indexed (reg a1) 4) (reg 1))
                        (*move (reg 2)(reg a1))
                        (ld!.l (indexed (reg a1) 4) (reg 2))
                        (lt!.d (reg 1) (reg 2))
                        (*move (fluid t) (reg 1))
                        (brs!.t  *donefgreaterp*)
                        (mov!.w (reg nil) (reg 1))
                        *donefgreaterp*))
 

(put 'wdivide 'opencode 
      % divide; return remainder in fluid
      '(  
        (mov!.w (reg 1) (reg 4)) 
        (div!.w (reg 2) (reg 1)) 
           % now compute remainder 
        (mul!.w (reg 1) (reg 2)) 
        (sub!.w (reg 2) (reg 4)) 
        (*move (reg 4) ($FLUID *second-value*)) 
     )) 
  

%------------------------------------------------------------------------
%
%   open codes for cons, xcons, ncons
(put 'cons 'opencode
      '((*move (fluid heaplast) (reg a1))
        (*move (fluid heaptrapbound)(reg a4))
        (*move (reg 1) (indexed (reg a1) 0) )
        (*move (reg 2) (indexed (reg a1) 4) )
        (*move (reg a1) (reg 1))
        (*wplus2 (reg a1) 8)
        (*move (reg a1) (fluid heaplast))
        (*wor (reg 1)(lshift pair-tag infbitlength))
        (*jumpwlessp (labelgen templabel) (reg a1)(reg a4)) 
        (*link !%reclaim-1 expr 1)  % is transparent wrt reg1
        (labelref templabel)))
        
(put 'xcons 'opencode
      '((*move (fluid heaplast) (reg a1)) 
        (*move (fluid heaptrapbound)(reg a4)) 
        (*move (reg 2) (indexed (reg a1) 0) ) 
        (*move (reg 1) (indexed (reg a1) 4) ) 
        (*move (reg a1) (reg 1)) 
        (*wplus2 (reg a1) 8) 
        (*move (reg a1) (fluid heaplast)) 
        (*mkitem (reg 1)(wconst pair-tag)) 
        (*jumpwlessp (labelgen templabel) (reg a1)(reg a4)) 
        (*link !%reclaim-1 expr 1) 
        (labelref templabel)))
 
(put 'ncons 'opencode
      '((*move (fluid heaplast) (reg a1)) 
        (*move (fluid heaptrapbound)(reg a4)) 
        (*move (reg 1) (indexed (reg a1) 0) ) 
        (*move (reg nil) (indexed (reg a1) 4) ) 
        (*move (reg a1) (reg 1)) 
        (*wplus2 (reg a1) 8) 
        (*move (reg a1) (fluid heaplast)) 
        (*mkitem (reg 1)(wconst pair-tag)) 
        (*jumpwlessp (labelgen templabel) (reg a1)(reg a4)) 
        (*link !%reclaim-1 expr 1) 
        (labelref templabel)))
