%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC/MULS.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for CONVEX processor
% Author:       H. Melenk
% Created:      10-Sept-90
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the 68020 processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
% They map the 2*30 bit operations needed by BIG to the 2*32 bit operations
% by shifting etc.

(fluid '(*second-value*))

(put 'wtimesdouble 'opencode
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The high order bits are placed in fluid variable.
     '((cvtw!.l (reg 1)(reg 1)) 
       (cvtw!.l (reg 2)(reg 2))
       (mul!.l (reg 2)(reg 1)) 
      % now we have 32 low bits in REG1, 32 high bits in REG3
       (mov!.l (reg 1) (reg 3))
       (*wand (reg 1) (wconst 16#3fffffff))
       (shf (wconst -30) (reg 3)0)
       (*MOVE (reg 3)($FLUID *second-value*))
     ))

(put 'wxtimes2 'opencode % different version for $pxu/mbarith
     '((cvtw!.l (reg 1)(reg 1))
       (cvtw!.l (reg 2)(reg 2))
       (mul!.l (reg 2)(reg 1))
       (mov!.l (reg 1)(reg 3))
       (shf (wconst -32) (reg 3))
       (*MOVE (reg 3)($FLUID *second-value*))
      ))

(put 'wquotientdouble 'opencode
      % called with a double length number in params 1 and 2
      % and a single length number in par 3.
      % Result is the single length quotient.
      % the remainder is placed in a fluid variable.
      '(  % adjusting the words first
        (cvtw!.l (reg 1)(reg 1))
        (shf (wconst 30)(reg 1))
        (cvtw!.l (reg 2)(reg 2))
        (add!.l (reg 2) (reg 1))
        (mov!.l (reg 1) (reg 4))
        (cvtw!.l (reg 3)(reg 3))
        (div!.l (reg 3)(reg 1))
           % now compute remainder
        (mul!.l (reg 1) (reg 3))
        (sub!.l (reg 3) (reg 4))
        (*move (reg 4) ($FLUID *second-value*))
     ))


(put 'addcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WPLUS2 (reg 1)(reg 2))
         (*MOVE (reg 1)(reg 3))
         (*WSHIFT (reg 3) -30)
         (*MOVE (reg 3) ($FLUID carry*))
         (*wand (reg 1) (wconst 16#3fffffff))))

(put 'subcarry 'opencode
       '(
         (*MOVE ($FLUID carry*) (reg 2))
         (*WDIFFERENCE (reg 1)(reg 2))
         (*MOVE (reg 1) (reg 2))
         (*WSHIFT (reg 2) -31)
         (*wand (reg 2) 1)
         (*MOVE (reg 2) ($FLUID carry*))
         (*wand (reg 1) (wconst 16#3fffffff))))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   functions supporting word arithmetic with bbase*
%   bbase* is too large to be stored in a fluid cell;
%   so we perform operations with bbase* by shift and mask
%   instructions.
 
%   get the value for bbase* into a register
(put 'bbase** 'opencode
      '((*MOVE 1 (reg 1))
        (*WSHIFT (reg 1) 30)))
 
%   get the value for logicalbits* into a register     
(put 'logicalbits** 'opencode
      '((*MOVE -1 (reg 1))
        (*WSHIFT (reg 1) -2)))

%   calculate (remainder x bbase*)
(put 'remainder-bbase 'opencode
      '((*wand (reg 1) (wconst 16#3fffffff))))

%   calculate (quotient x bbase*)
(put 'quotient-bbase 'opencode
      '((*WSHIFT (reg 1) -30)))

(put 'bbase** 'destroys '((reg 1))) 
(put 'logicalbits** 'destroys '((reg 1)))  
(put 'remainder-bbase 'destroys '((reg 1)))  
(put 'quotient-bbase 'destroys '((reg 1)))  

 

