%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         CRAY-LAP.SL
% Description:
% Author:       M. L. Griss & Robert Kessler, Utah Pass Project
% Created:      18-Nov-82
% Modified:     13-Dec-84 07:00:03 (Robert Kessler)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, University of Utah, all rights reserved
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin for modifications
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 26-Juny-87    H. Melenk, W. Neun,ZIB Berlin
%  Fixed horrible error in depositstring (with enlarging fasl space)
% 3-November-86    H. Melenk, W. Neun,ZIB Berlin
%  installed alignment of function to instruction buffer boundary
%  divided cray-lap into cray-lap and cray-lapopt
%  removed lap... arithmetic in favor of w... arithmetic
% 18-September-86    H. Melenk, W. Neun,ZIB Berlin
%  lambind and progbind produce inlinecode now.
%  therefore deleted them in links-present definition.
%
% 11-April-86 H. Melenk ZIB Berlin
%           added peephole optimizer on instruction base. First
%   application area: instruction schedule for load operations
%   (very Cray specific)
% 05/09/85 - Kessler Added check for links in code.
% 02/11/85 - Kessler, removed mapobl function, made a real function.
% 12/27/84 - Kessler
%  Added parcellabel operandvaluefunction.
% 11-Dec-84 15:34:18 (Robert Kessler)
%  Changed order in forwardinternalreferences* so name is first.
% 11-Dec-84 10:05:47 (Robert Kessler)
%  Fixed Systemfaslfixup so it works correctly with the CRAY.  THIS IS
%  DIFFERENT FROM THE VERSIONS ON OTHER MACHINES, IN THAT IF AN
%  INTERNAL FUNCTION REFERENCE IS FOUND THAT IS NOT DEFINED, IT CAUSES
%  AN ERROR.
% 11-Dec-84 06:45:58 (Robert Kessler)
%  Changed putentry call to use a word offset, instead of parcel.ki
% 10-Dec-84 08:53:20 (Robert Kessler)
%  Fixed bug in forceupoffset which used just the codesize, instead of
%  using the currentoffset*.  This is only a problem for fasld code,
%  since the currentoffset* will be zero in new in core compilations.
% 27-Nov-83 11:50:48 (Robert Kessler)
%  Added new calls to add-next-bit, and removed all updatebittables,
%  except for the ones that are used to update based upon increasing
%  the code address.
%  Changed lap, so it checks that currentoffset starts on a word
%  boundary when we begin each new piece of lap code.
%  Added auto-update-bittable and add-next-bit functions.
%  <PSL.KERNEL.CRAY>CRAY-LAP.RED.35, 29-May-84 07:55:33, Edit by KESSLER
%  Changed occurances of atsoc to assoc, so kludgy lapout/in will work with
% labels that are really strings.
% Kessler - 3/22/84 - Changed float laplength to 8.
%  <PSL.KERNEL.CRAY>CRAY-LAP.RED.32,  7-Sep-83 17:26:50, Edit by KESSLER
%  Change DepositItem so it does a word labelvalue instead of parcel.
%  <PSL.KERNEL.CRAY>CRAY-LAP.RED.29,  6-Sep-83 07:49:58, Edit by KESSLER
%  Labelp updated to know about strings.
%  <PSL.KERNEL.CRAY>CRAY-LAP.RED.24, 31-Aug-83 09:03:33, Edit by KESSLER
%  GTBPS returns word address and wants words to allocate.  So had
%   to force words for size, and turn into a parcel address for later use.
%   Deleted definitions of lapland, laplsh and friends, already defined in
%   cray-forms.
% 8/16/83 - Kessler
%  Changed forceup in computing the lengths in pass1, so it will check
%   see if it has to update the label offsets.
%  Changed DepositInstruction, so it will also have opcodeFexprFunction
%   for those opcodes that have a variable number of arguments.
% 8/15/83 - Kessler
%  Bug in depositword, it updates currentoffset and so does crayputword.
%  Edit to Lap, changed to printout the starting base address.
%Revisions: With Bob and Dana
%      Fix LAPLENGTHs
%      Simulate CRAY-BPS in BPS
%
% Based on 20 LAP
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(imports '(cray-lapopt))
 
(compiletime(progn
(put 'reloc_word          'const 1)
(put 'reloc_id_number     'const 2)
(put 'escape_to_second    'const 3)
(put 'parcel_label        'const 0)
(put 'word_label          'const 1)
(put 'reloc_value_cell    'const 2)
(put 'reloc_function_cell 'const 3)
nil))
 
(fluid '(labeloffsets* currentoffset* codesize* codebase* entries*
         forwardinternalreferences*
         newbittableentry* lapreturnvalue*
         *nextbit* NAlloc*
         *writingfaslfile initoffset* *pgwd *pwrds))
 
%/ Initially work in parcel address, but may need to keep
%/ both word and parcel attribute
 
(commentoutcode
(flag '(saveentry defineentries depositinstruction
       opcodevalue operandvalue depositword depositwordexpression
       deposithalfwords labelvalue deposititem deposithalfwordidnumber
       findlabels onelaplength makerelocinf makerelocword)
     'internalfunction))
 
(compiletime (flag '(adjustlabels) 'internalfunction))
 
(de labelp (x)
    (or (atom x) (stringp x) (eqcar x 'label)))
 
(de forceup (p)
 % Get Parcel address to next word
 (wshift (wshift (wplus2 p 3) -2) 2 ))
 
(de forcedown (p)
 (wshift (wshift p -2) 2 ))
 
% New verison of lap that checks for the presence of any *link cmacros.
(fluid '(&links-present&))
 
(de lap (u)
  (prog (lapreturnvalue* labeloffsets* entries* adjusted gaplength)
    (cond ((not *writingfaslfile)
         (setq currentoffset* 0))
          (t (while (not (wordboundaryp currentoffset*))
             (progn
             % Set the bit table to ignore the parcels.
              (updatebittable 1 0)
              (setq currentoffset* (wplus2 currentoffset* 1))))))
    (setf &links-present& (or (look-for-links u)
%     (assoc '*lambind u) (assoc '*progbind u)
      (assoc '*Id-apply u)
      (assoc '*call u)))
    (setq u (pass1lap u))
    (findlabels u)
    (cond (*pgwd (progn (terpri) (terpri))))
    (cond (*pgwd
         (foreach x in u do
                (cond ((atom x) (prin2 x))
                    (t (printf "            %p%n" x))))))
    (cond (*pgwd (progn (terpri) (terpri))))
    (if *writingfaslfile
         (progn
             (setq adjusted (wlshift (wrshift
                                            (wplus2 currentoffset* 127)
                                  7)        7))
             (when (wgeq (wplus2    codesize*
                                    currentoffset*)  adjusted)
                   (setq gaplength 0)
                   (while (wlessp currentoffset* adjusted)
                     (progn
                        % Set the bit table to ignore the parcels.
                      (updatebittable 1 0)
                      (setq gaplength (wplus2 gaplength 1))
                      (setq currentoffset* (wplus2 currentoffset* 1))))
                   (adjustlabels gaplength)
          )  )
         (setq codebase* (wtimes2 (gtbps (wquotient
                                (forceup codesize*) 4)) 4)))
    % Need to work in PARCELS, GtBPS in WORDS
    (cond (*pgwd
         (prog (outputbase*)
           (setq outputbase* 8)
           (printf "%nBegining Code Base = %p%n"(wquotient codebase* 4)))))
 
    (foreach x in u do
           (cond ((not (labelp x))
                (cond ((equal (first x) '*entry) (saveentry x)) 
                       (t (depositinstruction x))))))
    (defineentries)
    (cond ((and (not *writingfaslfile) *pwrds)
         (errorprintf "*** %p: base %o, length %d words"
                  (foreach x in entries* collect (first (car x)))
                 (wquotient codebase* 4)(wquotient (forceup codesize*) 4))))
    (cond (lapreturnvalue*
         (return (mkcode (wquotient lapreturnvalue* 4)))))))     % MKcode in WOR
 
(de adjustlabels(diff)
       (setq labeloffsets*
           (mapcar labeloffsets*
                  (function (lambda (x)
                             (cons (car x) (wplus2 (cdr x) diff)))))))
 
(de look-for-links (u)
       (cond ((null u) nil)
             ((atom (car u)) (look-for-links (cdr u)))
             ((eq (caar u) '*link)
                  (cond
                    ((eq  (cadar u) 'catchsetup) t) % is opencode
                         % but problems will arise in throw if no
                         % allocation takes place
                    ((get (cadar u) 'opencode) (look-for-links (cdr u)))
                    (t t)))
             (t (look-for-links (cdr u)))
)      ) 
 
 
(de saveentry (x)
  (cond ((equal (second x) '***code**pointer***)
       (setq lapreturnvalue*           % Magic token that tells LAP to return
         (cond (*writingfaslfile currentoffset*)
             (t       %  a code pointer
              (wplus2 codebase* currentoffset*)))))
      ((not *writingfaslfile)
       (progn
        (setq entries* (cons (cons (rest x) currentoffset*) entries*))
        (cond ((not lapreturnvalue*)
             (setq lapreturnvalue*
               (wplus2 codebase* currentoffset*))))))
      ((equal (second x) '**fasl**initcode**)
       (setq initoffset* currentoffset*))
      ((flagp (second x) 'internalfunction)
       (put (second x) 'internalentryoffset currentoffset*))
      (t (progn
          (findidnumber (second x))
          (dfprintfasl (list 'putentry (mkquote (second x))
                         (mkquote (third x))
                         (parceltoword currentoffset*)))))))
 
(de defineentries nil
    (foreach x in entries* do
      (putd (first (car x))
            (second (car x))
            (mkcode (wquotient (wplus2 codebase* (cdr x)) 4)))))
 
(de depositinstruction (x)
  %
  % Legal forms are:
  %  (special_form . any)
  %
  (prog (y)
    (return
     (cond ((setq y (get (first x) 'instructiondepositfunction))
              (apply y (list x)))
           ((setq y (get (first x) 'opcodefunction))
              (apply y (cdr x)))
           ((setq y (get (first x) 'opcodefexprfunction))
              (apply y (list (cdr x))))
           (t (stderror (bldmsg "%p is not a legal OP form%n" x)))))))
 
(de operandvalue (u)
  %
  % Legal forms are:
  % number
  % other atom (label)
  % (special . any)      fluid, global, etc.
  % (reg name/number)
  %
  (prog (x)
    (return (cond ((setq x (get (first u) 'operandvaluefunction))
                       (apply x (list u)))
                  ((setq x (wconstevaluable u)) (operandvalue x))
                  (t (stderror (bldmsg "Unknown operand %r" u)))))))
 
 
(de binaryoperand (u)
  %
  % (op x x) can occur in expressions
  %
  (prog (x)
    (return
     (cond ((setq x (wconstevaluable u)) x)
         (t (progn
             (setq x (cond ((getd (first u)) (first u))
                       (t (get (first u) 'dofn))))
             (setq u (rest u))
             (cond ((numberp (first u))
                    (apply x (list (first u) (labelvalue (second u)))))
                   ((numberp (second u))
                    (apply x (list (labelvalue (first u)) (second u))))
                   (t (stderror
                   (bldmsg "Expression too complicated in LAP %r" u))))))))))
 
 
% Add others to this list if they arise
 
(put 'difference 'operandvaluefunction 'binaryoperand)
(put 'plus2     'operandvaluefunction 'binaryoperand)
 
%(de registeroperand (u)
%  (prog (regname v)
%    (setq regname (second u))
%    (return
%     (cond ((posintp regname)
%                (list 'reg (indx numericregisternames regname)))
%           ((setq v (get regname 'registerrename)) (list 'reg v))
%           ((or (aregp u) (bregp u) (sregp u) (tregp u) (vregp u))
%                       u)
%           (t (stderror (bldmsg "Unknown register %r" u)))))))
%
%(put 'reg 'operandvaluefunction 'registeroperand)
 
(de immediateoperand (u)
  (operandvalue (second u)))          % immediate does nothing on the PDP10
 
(put 'immediate 'operandvaluefunction 'immediateoperand)
 
(de lapvaluecell (u)
  (valuecelllocation (second u)))
 
(deflist '((fluid lapvaluecell)
           ($fluid lapvaluecell)
           (global lapvaluecell)
           ($global lapvaluecell)) 'operandvaluefunction)
 
(de lapentry (u)
  (functioncelllocation (second u)))
 
(put 'entry 'operandvaluefunction 'lapentry)
 
%(de lapinternalentry (u)
%  (prog (x)
%    (setq u (second u))
%    (setq newbittableentry* (const reloc_halfword))
%    (return
%     (cond
%      % Do we know about the internal entry yet or not?
%      % We need to set some bits!!
%      ((setq x (assoc u labeloffsets*))
%
%       (progn (setq x (cdr x))
%            (cond (*writingfaslfile
%                 (add-next-bit (const escape_to_second))
%                 (add-next-bit (const parcel_label))
%                 x)
%                (t (lapplus2 codebase* x)))))
%      % Oh well, we will need to fix things up later.
%      (t (progn
%        (cond ((not *writingfaslfile)
%             (functioncelllocation u))
%            ((setq x (get u 'internalentryoffset))
%             (add-next-bit (const escape_to_second))
%             (add-next-bit (const parcel_label))
%             x)
%            % Aha, we don't yet know what the address is, so we
%            % need to set things to fix it up later.
%            (t (progn
%                (add-next-bit (const escape_to_second))
%                (add-next-bit (const parcel_label))
%                (setq forwardinternalreferences*
%                  (cons (cons currentoffset* u)
%                      forwardinternalreferences*))
%                0)))))))))
%
%(put 'internalentry 'operandvaluefunction 'lapinternalentry)
 
(de depositwordblock (x)
    (foreach y in (cdr x) do (depositwordexpression y)))
 
(put 'fullword 'instructiondepositfunction 'depositwordblock)
 
% (de deposithalfwordblock (x)
% (prog (l r)
%     (setq x (rest x))
%     (while (not (null x))
%     (progn (setq l (first x))
%       (setq x (rest x))
%       (cond ((null x)
%          (setq r 0)) (t
%
%       (progn (setq r (first x))
%           (setq x (rest x)))))
%       (deposithalfwords l r)))))
%
%
% (put 'halfword 'instructiondepositfunction 'deposithalfwordblock)
 
(de depositstring (x)
(prog (y z)
  % May have to do some fiddle for CROSS assembly into parcels
    (setq z (second x))
    (setq y (strpack (strlen (strinf z))))
    (for (from i 1 y 1) (do (depositword (sys2int (getmem
                                                    (aplus2 z i))))))))
 
(put 'string 'instructiondepositfunction 'depositstring)
 
(de depositfloat (x)                    % this will not work in cross-assembly
(progn (setq x (second x))              % don't need to strip tag on PDP10
    (depositword (sys2int (floathighorder x)))
    (depositword 0                          )))
%   (depositword (sys2int (floatloworder x)))))
 
(put 'float 'instructiondepositfunction 'depositfloat)
 
(de wordboundaryp (x)
  (equal (lapland x 3) 0))
 
% (de halfwordboundaryp (x)
%   (equal (lapland x 1) 0))
 
% Stick the word argument into memory on the proper boundary.
(de depositword (x)
  (progn 
   (while (not (wordboundaryp currentoffset*))
     (progn
      % Set the bit table to ignore the parcels.
      (updatebittable 1 0)
      (setq currentoffset* (wplus2 currentoffset* 1))))
   % Since crayputword updates the currentoffset, we must do the
   % updatebittable first.
   % (updatebittable 4 0), not necessary, since it will default to 0.
   (crayputword (CalculateAddress) x)
   nil)) 
 
% Only limited expressions now handled
(de depositwordexpression (x)
% Not CRAY VALID
  (prog (y)
    % First we go to the word boundary.
    (while (not (wordboundaryp currentoffset*))
      (progn
       (updatebittable 1 0)
       (setq currentoffset* (wplus2 currentoffset* 1))))
    (return
     (cond ((fixp x) (depositword (int2sys x)))
         % When will we ever generate a label?  The only time this
         % gets called is when we have a fullword in the code.
         % Thus, it is quite possible that we never stick a label in
         % the fullword??
         % Oh, well, we will relocate the word to the base.
         ((labelp x)
          (progn
           (add-next-bit (const reloc_word))
           % (updatebittable 4 (const reloc_word)), don't need,
           % since rest will be zero.
           (crayputword (CalculateAddress) (labelvalue x))
           nil))
         %      CurrentOffset!* := LapPlus2(CurrentOffset!*,4)
         ((equal (first x) 'mkitem) (deposititem (second x) (third x)))
         ((equal (first x) 'fieldpointer)
          (depositfieldpointer (second x) (third x) (fourth x)))
         ((setq y (wconstevaluable x)) (depositword (int2sys y)))
         (t (stderror (bldmsg "Expression too complicated %r" x)))))))
 
% (de deposithalfwords (l r)
% % No quite CRAY valid
% (prog (y)
%     (cond ((not (or (fixp l) (setq l (wconstevaluable l))))
%       (stderror "Left half too complex")))
%     (cond ((and (pairp r) (equal (first r) 'idloc))
%       (deposithalfwordidnumber l (second r)))
%     ((setq y (wconstevaluable r)) (depositword (laplor (laplsh l 18) y))) (t
%     (stderror (bldmsg "Halfword expression too complicated %r" r))))))
 
(de labelvalue (u)
  (prog (v)
    (return
     (cond ((codep u) (wtimes2 4 (inf u)))       % Word to Parcel
         ((setq v (assoc u labeloffsets*))
          (progn (setq v (cdr v))
               (cond (*writingfaslfile
                    % Don't think the newbit stuff is necessary.
                    % (progn
                    % (setq newbittableentry* (const reloc_halfword))
                    % v))
                    v)
                   (t (wplus2 codebase* v)))))
         (t (stderror (bldmsg "Unknown label %r in LAP" u)))))))
 
 
(de deposititem (tagpart infpart)
  (cond ((not *writingfaslfile)
       (depositword (sys2int
                   (mkitem tagpart
                         (cond ((labelp infpart)
                              (wquotient (labelvalue infpart) 4))
                             ((equal (first infpart) 'idloc)
                              (idinf (second infpart)))
                             (t (stderror
                               (bldmsg "Unknown inf in MkItem %r"
                                     infpart))))))))
      (t (progn
          (cond ((labelp infpart)
               % The following could be trouble, if this is to be
               % relocated as a code pointer or something.
               % (updatebittable 4 (const reloc_word)), Don't need
               (add-next-bit (const reloc_word))
               (crayputword (CalculateAddress)
                        % RELOC_CODE_OFFSET = 0
                        (sys2int
                         (mkitem tagpart
                               (parceltoword (labelvalue infpart))))))
              ((equal (first infpart) 'idloc)
               % This is if we are making something with an idloc
               % in it.
               (add-next-bit 0)
               (add-next-bit 0)
               (add-next-bit (const reloc_id_number))  % Escape to second
               (add-next-bit 0)   % And relocate simple id.
               (crayputword (CalculateAddress)
                        (sys2int (mkitem tagpart
                               (findidnumber (second
                                          infpart))))))
              (t (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))
          %      CurrentOffset!* := LapPlus2(CurrentOffset!*,4);
          % (updatebittable 1 (const reloc_inf))
          ))))
 
% (de deposithalfwordidnumber (lhs x)
%     (cond ((or (not *writingfaslfile) (ileq (idinf x) 128))
%       (depositword (laplor (laplsh lhs 18) (idinf x)))) (t
%
%     (progn (crayputword (laplus2 codebase* currentoffset*)
%              (laplor (laplsh lhs 18))
%             (makerelochalfword (const reloc_id_number) (findidnumber x)))
% %      CurrentOffset!* := LapAdd1 CurrentOffset!*;
%       (updatebittable 1 (const reloc_halfword))))))
 
% Purpose is to resolve any internal references.  The
% forwardinternalreferences* variable contains an offset and a name of
% an id that needs to be fixed up.
(de systemfaslfixup nil
  (progn 
   (while (not (null forwardinternalreferences*))
     (prog
       (Internal-Offset Internal-Name)
       (setf Internal-Offset (wplus2 Codebase*
                           (cdr (first forwardinternalreferences*))))
       (setf Internal-Name (car (first forwardinternalreferences*)))
       (cond ((get Internal-Name 'internalentryoffset)
            (put-2-parcel
             Internal-Offset
             (wplus2 (get Internal-Name 'internalentryoffset)
                     (get-2-parcel Internal-Offset))))
           (t (stderror
             (bldmsg
              "***** %r not defined in this module.  May not be internal."
              Internal-Name))))
       (setq forwardinternalreferences* (cdr forwardinternalreferences*))))
   (mapobl 'reminternalentryoffset)))
 
% Kludge since the cross assembler can't handle internal functions.
(de reminternalentryoffset (x)
 (remprop x 'internalentryoffset))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Code needed to update the bittable properly.  The
% auto-update-bittable will update according to the value in
% *nextbit*, defaulting to 0 if nothing there.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(de auto-update-bittable ()
  (cond (*writingfaslfile
       (cond (*nextbit* (updatebittable 1 (car *nextbit*))
                    (setf *nextbit* (cdr *nextbit*)))
             (t (updatebittable 1 0))))))
 
% This function will add another constant to the *nextbit* list
% (careful to put it at the end instead of the front).
(de add-next-bit (bit)
  (setf *nextbit* (append *nextbit* (list bit))))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% PASS1 LABEL stuff
% Increment Codesize!* by parcel size of each instruc, define lables
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(lapcodelist*))
 
(de findlabels (lapcodelist*)
(progn (setq codesize* 0)
    (foreach x in lapcodelist* do
    (progn (setq x (onelaplength x))
       (setq codesize* (wplus2 codesize* x))))))
 
(de onelaplength (u)
% determine size of each instruc
  (prog (x)
    (return
     (cond
      ((atom u)
 
       % Might be better to "defer" labels on a stack
       % then define at next instruction once PARCEL/WORD forceup done
       % and PARCEL/WORD attribute determined... later
 
       (progn
      (setq labeloffsets*
        (cons (cons u (wplus2 currentoffset* codesize*))
            labeloffsets*)) 0))
      ((setq x (get (car u) 'laplength))
       (cond ((posintp x) x)
           (t (apply x (list u)))))
      (t 
       % minor klugde for long constants
       %/ Doesnt seem needed for CRAY
       (progn
      (cond
       ((and (equal (length u) 3) (fixnp (setq x (third u)))
             (not (immediatep x)))
        (prog (y)
          (rplaca (rest (rest u)) (setq y (stringgensym)))
          (nconc lapcodelist* (list y (list 'fullword x))))))
      1))))))
 
    % default 1 parcel instructions
 
% Do a force up while computing the length of the code.  We must also update
%  the label table if the forceup screwed up the offset.
(de forceupoffset (oldoffset)
  % Get Parcel address to next word
  (prog (labels newoffset compare-offset)
    (setq newoffset (wtimes2 4 (wquotient (wplus2 oldoffset 3) 4)))
    (cond ((equal newoffset oldoffset) (return newoffset)))
    % Ok, we have done a forceup, so we must modify all of the labels in the
    %  labeloffsets* label list.
    (setq labels labeloffsets*)
    (while (pairp labels)
      (progn
       (setq currentlabel (car labels))
       (setq labels (cdr labels))
       (setq compare-offset (wplus2 currentoffset* codesize*))
       (cond ((wlessp (cdr currentlabel) compare-offset)
            (setq labels nil))
           ((equal (cdr currentlabel) compare-offset)
            (rplacd currentlabel (wplus2 currentoffset* newoffset))))))
    (return newoffset)))
 
 
(deflist '((*entry   lapentrylength)
           (float    8)
           (string   lapstringlength)
           (fullword lapwordlength)
           (halfword laphalfwordlength)
           (byte     lapbytelength)) 'laplength)
 
(de lapentrylength (u)
% Make a label for each (!*ENTRY name type nargs)
(progn (setq codesize* (forceupoffset codesize*))       % Start proc
   (setq labeloffsets* (cons
                  (cons (second u) (wplus2 currentoffset* codesize*))
                                   labeloffsets*)) 0))
 
(de lapstringlength (u)
(progn (setq codesize* (forceupoffset codesize*))
   (wtimes2 4 (strpack (strlen (strinf (second u)))))))
 
(de lapwordlength (u)
(progn (setq codesize* (forceupoffset codesize*))
    (wtimes2 4 (length (rest u)))))
 
%(de laphalfwordlength (u)
%(progn (setq codesize* (forceupoffset codesize*))
%    (times 4 (laplsh (lapadd1 (length (rest u))) -1))))
 
 
(de lapbytelength (u)
(progn (setq codesize* (forceupoffset codesize*))
    (wtimes2 4 (strpack (length (rest u))))))
 
% May need to handle LAPALIGNMENT as well as LENGTH and OPCODE
 
% ExtraRegLocation is in CRAY-FORMS
 
(put 'extrareg 'operandvaluefunction 'extrareglocation)
 
%(de makerelocword (reloctag relocinf) (wplus2 (wshift
%    reloctag 34) (field relocinf 2 34)))
 
%(de makerelocinf (reloctag relocinf) (wplus2 (wshift
%    reloctag 16) (field relocinf 20 16)))
 
%(de makerelochalfword (reloctag relocinf) (wplus2 (wshift
%    reloctag 16) (field relocinf 20 16)))
 
% Parcel extracts the warray's address and turns into a parcel.
(de parcellabel (u)
     (cond ((labelvalue (cadr u))
            (labelvalue (cadr u)))))
 
(put 'parcel 'operandvaluefunction 'parcellabel)
 
