%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  File: CRAY-LAPOPT-4
%
%  Description: Cray Code Optimization
%               Full instruction scheduling, instruction data
%
%  Author:      H. Melenk, ZIB Berlin
%  Date:        16 Dec 1986
%
% Copyright (c) 1987,  Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%% instruction scheduler for Cray PSL
%%
%% instruction data
 
(compiletime
  (load  f-evectors) (on fast-integers)
 )
 
(compiletime (progn
 
%(defstruct (inst :print-function print-node)
%
% % static description:
%    nbr                  % original number in the code sequence
%    instruction          % original code of the instruction
%    inputs
%    outputs
%    unit                 % unit occupied by the instruction
%    time                 % time needed
%
% % network slots:
%    (input-predecessors nil)
%    (output-predecessors nil)
%    (input-successors nil)
%    (output-successors nil)
%
% % working slots:
%    (limit  0)           % earliest time to start
%    (weight 0)           % "importance" of the instruction
%)
  (ds inst-nbr ($$x) (egetv $$x 1))
  (ds setf-inst-nbr ($$x $$v) (eputv $$x 1 $$v))
  (put 'inst-nbr 'assign-op 'setf-inst-nbr)
 
  (ds inst-instruction ($$x) (egetv $$x 2))
  (ds setf-inst-instruction ($$x $$v) (eputv $$x 2 $$v))
  (put 'inst-instruction 'assign-op 'setf-inst-instruction)
 
  (ds inst-inputs ($$x) (egetv $$x 3))
  (ds setf-inst-inputs ($$x $$v) (eputv $$x 3 $$v))
  (put 'inst-inputs 'assign-op 'setf-inst-inputs)
 
  (ds inst-outputs ($$x) (egetv $$x 4))
  (ds setf-inst-outputs ($$x $$v) (eputv $$x 4 $$v))
  (put 'inst-outputs 'assign-op 'setf-inst-outputs)
 
  (ds inst-unit ($$x) (egetv $$x 5))
  (ds setf-inst-unit ($$x $$v) (eputv $$x 5 $$v))
  (put 'inst-unit 'assign-op 'setf-inst-unit)
 
  (ds inst-time ($$x) (egetv $$x 6))
  (ds setf-inst-time ($$x $$v) (eputv $$x 6 $$v))
  (put 'inst-time 'assign-op 'setf-inst-time)
 
  (ds inst-input-predecessors ($$x) (egetv $$x 7))
  (ds setf-inst-input-predecessors ($$x $$v) (eputv $$x 7 $$v))
  (put 'inst-input-predecessors 'assign-op 'setf-inst-input-predecessors)
 
  (ds inst-output-predecessors ($$x) (egetv $$x 8))
  (ds setf-inst-output-predecessors ($$x $$v) (eputv $$x 8 $$v))
  (put 'inst-output-predecessors 'assign-op 'setf-inst-output-predecessors)
 
  (ds inst-input-successors ($$x) (egetv $$x 9))
  (ds setf-inst-input-successors ($$x $$v) (eputv $$x 9 $$v))
  (put 'inst-input-successors 'assign-op 'setf-inst-input-successors)
 
  (ds inst-output-successors ($$x) (egetv $$x 10))
  (ds setf-inst-output-successors ($$x $$v) (eputv $$x 10 $$v))
  (put 'inst-output-successors 'assign-op 'setf-inst-output-successors)
 
  (ds inst-limit ($$x) (egetv $$x 11))
  (ds setf-inst-limit ($$x $$v) (eputv $$x 11 $$v))
  (put 'inst-limit 'assign-op 'setf-inst-limit)
 
  (ds inst-weight ($$x) (egetv $$x 12))
  (ds setf-inst-weight ($$x $$v) (eputv $$x 12 $$v))
  (put 'inst-weight 'assign-op 'setf-inst-weight)
 
  (ds commonlispmapc (a b) (mapc b a))
  (ds princ (x) (prin2 x))
  (ds consp (x) (pairp x))
 
  (defmacro  defun (u . v) `(de ,u . ,v))
  (defmacro  funcall (u . v) `(idapply ,u (list . ,v)))
  (defmacro  loop    (u . v) `(while t ,u . ,v))
 
  (ds commonlisppush($a $b)
   (setf $b (cons $a $b)))
 
) )
 
  %% calculation ot the weight of a given instruction. Cray
  %% formula:
  %%  execution time + weight of dependent instructions
  %%
(defun CalculateWeight (struct)
   (prog (x p deps)
            % collect all dependent instructions
        (setq deps
           (collectDependentInstructions
                (inst-input-successors struct)
                deps))
        (setq deps
           (collectDependentInstructions
                (inst-output-successors struct)
                deps))
        (setq x (+ (inst-time struct)   % start with operation length
                   (inst-weight struct))) % plus orig. weight
 
        (loop
             (when (or (null deps) (> x 100000)) (return nil))
             (setq x (+ (inst-time (car deps)) x))
             (setq x (+ (inst-weight (car deps)) x))
             (setq deps (cdr deps))   )
 
        (setf (inst-weight struct) x) ))
 
(defun collectDependentInstructions (instrs deps)
   (when instrs
     (CommonLispMapc
        #'(lambda(x)
            (when (not (memq x deps))
                  (setq deps (cons x deps))
                  (setq deps
                      (collectDependentInstructions
                          (inst-input-successors x)
                          deps))
                  (setq deps
                     (collectDependentInstructions
                          (inst-output-successors x)
                          deps))
          ) )
          instrs))
   deps) 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Scalar Instructions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% note that scalar instructions do not reserve units
% so all units are NIL
 
  %% standard form function: 1st parameter output,
  %%                         rest of parameters input
(defun StdFormFn (inst struct)
    (setf (inst-inputs struct)(append (cddr inst) nil))
    (setf (inst-outputs struct)(list (cadr inst)))
    t)
 
  %% for instructions available for A-regs and S-regs
  %%             distinguish the time
(defun SAFormFn (inst struct)
  (let ((mod (sregp (cadr inst))))
   (setf (inst-time struct)
         (If mod
             (car (inst-time struct))
             (cdr (inst-time struct))))
   (stdFormFn inst struct)))
 
 
(defun strip-areg (u)
   (let (x)
    (cond ((not (consp u)) nil)
          ((aregp u) (list u))
          ((setq x (strip-areg (car u))) x)
          (t (strip-areg (cdr u))) )))
 
  %%% special cases
   %% MOVE
(defun MoveFormFn (inst struct)
  (let ((target (cadr inst)) (source (caddr inst)))
    (if   (sregp target)
          (cond ((aregp source)
                 (setf (inst-time struct) 2))          % special timing
                ((and (intp source) (not (eq source -1))
                      (not (eq source 1))(not (eq source 0)))
                 (setf (inst-time struct) 2))          % special timing
         %      ((equal target '(Reg S0))        %preference for A0/S0
         %       (setf (inst-weight struct) 5) )
                 )
          (cond %((equal target '(Reg A0))
                % (setf (inst-weight struct) 5) )
                ((and (intp source) (wgreaterp source 64))
                 (setf (inst-time struct) 2))))        % special timing
    (StdFormFn inst struct))) % do standard things
 
   %% LOAD
(defun loadFormFn (inst struct)
    (setf (inst-inputs struct)
            (append (strip-areg (cddr inst)) '(memory)))
    (setf (inst-weight struct) 8)
    (setf (inst-outputs struct) (list (cadr inst))))
 
   %% STORE
(defun storeFormFn (inst struct)
    (setf (inst-outputs struct) '(memory))
    (setf (inst-inputs struct) (cons (cadr inst)
                                     (strip-areg (cddr inst)))) )
 
   %% DOUBLE SHIFTS and MERGEs
(defun dshiftandmergeFormFn (inst struct)
    (setf (inst-outputs struct) (list (cadr inst) (cddr inst)))
    (setf (inst-inputs struct) (cdr inst)))
 
(fluid '(*AllTregs*))
 
   %% BLOCKLOAD
(defun BlockloadFormFn (inst struct)
    (when (not (equal (caddr inst) '(Reg Prevframesize)))
          (return nil))
    (setf (inst-time struct) 15)        % time consuming
    (setf (inst-weight struct) 100)     % start early
    (setf (inst-inputs struct)
            (list (cadr inst) '(reg A0) 'memory))
    (setf (inst-outputs struct) (append *allTregs*
                            '((reg t21) (reg t00) (reg T) (reg nil)))))
 
% blockload should not 'block' the return T/NIL operation
 
(commentoutcode
   %% can areise only inside alloc sequence; do not move there
   %% BLOCKSTORE
(defun BlockstoreFormFn (inst struct)
    (when (not (equal (caddr inst) '(Reg Prevframesize)))
          (return nil))
    (setf (inst-weight struct) 10000)       % as early as possible
    (setf (inst-inputs struct)
            (append *allTregs* (list (cadr inst) '(reg A0)) ))
    (setf (inst-outputs struct) '(memory)))
 )
 
(defun EriformFn (inst struct)
  (setf  (inst-inputs struct) (list '(reg 1)'(reg 2)'(reg 3) '(reg 4)
                                    'memory
                                    '(reg 5)'(reg s6)'(reg s7)))
  (setf  (inst-outputs struct) '(memory)))
 
(defun DriformFn (inst struct)
  (setf  (inst-outputs struct) '(memory))
  (setf  (inst-inputs struct) '(memory)))
 
(setq *allTregs*
   '((Reg ReturnAddr)(Reg PrevFrameSize)
     (Reg T30)(Reg T31)(Reg T32)(Reg T33)(Reg T34)(Reg T35)(Reg T36)(Reg T37)
     (Reg T40)(Reg T41)(Reg T42)(Reg T43)(Reg T44)(Reg T45)(Reg T46)(Reg T47)
     (Reg T50)(Reg T51)(Reg T52)(Reg T53)(Reg T54)(Reg T55)(Reg T56)(Reg T57)
     (Reg T60)(Reg T61)(Reg T62)(Reg T63)(Reg T64)(Reg T65)(Reg T66)(Reg T67)
     (Reg T70)
    ))
%%% INSTRUCTIONS
 
%% simple moves
 
(commonlispmapc
    #'(lambda(p)
          (setf (get p 'LapSchedUnit) nil)
          (setf (get p 'LapSchedTime) 1)
          (setf (get p 'LapSchedFormFn) 'MoveFormFn))
     '(move movesgn movefl sconst04 sconst06 moveneg))
 
%% simple logical instructions
(commonlispmapc
    #'(lambda(p)
          (setf (get p 'LapSchedUnit) 'nil)
          (setf (get p 'LapSchedTime) 1)
          (setf (get p 'LapSchedFormFn) 'StdFormFn))
     '(rmask lmask sbit cmpsbit and or not cmpand xor
       cmpxor))
 
%% simple shifts instructions
(commonlispmapc
    #'(lambda(p)
          (setf (get p 'LapSchedUnit) 'nil)
          (setf (get p 'LapSchedTime) 2)
          (setf (get p 'LapSchedFormFn) 'StdFormFn))
     '(lshift rshift))
 
%% double shifts instructions and merges
(commonlispmapc
    #'(lambda(p)
          (setf (get p 'LapSchedUnit) 'nil)
          (setf (get p 'LapSchedTime) 3)
          (setf (get p 'LapSchedFormFn) 'dshiftandmergeFormFn))
     '(dlshift drshift merge movsbit))
 
%% individuals
(commonlispmapc
    #'(lambda (p)
          (prog (u)
             (setq u (car p))
             (setf (get u 'LapSchedUnit) (cadr p))
             (setf (get u 'LapSchedTime) (caddr p))
             (setf (get u 'LapSchedFormFn) (cadddr p))
      )   )
      '( 
         (add nil (3 . 2) SAFormFn)
         (sub nil (3 . 2) SAFormFn)
         (mul nil  6  StdFormFn)
         (fadd nil  6  StdFormFn)
         (fsub nil 6  StdFormFn)
         (fmul nil 7  StdFormFn)
         (hmul nil 7  StdFormFn)
         (rmul nil 7  StdFormFn)
         (imul nil  7  StdFormFn)
         (hrecip nil 14  StdFormFn)
         (load nil 13 LoadFormFn)
         (store nil 1 StoreFormFn)
 %       (blksto nil 1 BlockStoreFormFn)
         (blklod nil 1 BlockLoadFormFn)
         (zeropop nil 3 StdFormFn)
         (onepop nil 3 StdFormFn)
         (paritypop nil 3 StdFormFn)
         (eri nil 1 EriFormFn)
         (dri nil 3 DriFormFn)
         (Dummy nil 4 StdFormFn) % dummy instructions to schedule S0/A0
)    )
