%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXC/i860-spec.sl
% Title:          Some special things for MIPS PSL compiler
% Author:         Winfried Neun , ZIB Berlin
% Created:        May 15, 1992
% Modified:
% Status:         Experimental
% Mode:           Lisp
%
% Copyright (c) 1992, Konrad-Zuse-Zentrum Berin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% a special Linke to save the NOP instruction after final Jump

(de *LinkE (DeAllocCount FunctionName FunctionType NumberOfArguments)
  (cond ((FlagP FunctionName 'ForeignFunction)
              `((*ForeignLink ,FunctionName ,FunctionType ,NumberOfArguments) 
                (*Exit 0)))
        ((get functionname 'exitopencode)
              `((*DeAlloc ,DeAllocCount)
                (*JCall ,FunctionName)))
        ((get functionname 'opencode)
              `((*Call ,FunctionName)
                (*exit ,DeAllocCount)))
        ((InternallyCallableP functionname)
              `((*dealloc ,Dealloccount)
		(br (InternalEntry ,functionname))
                (nop)))
        (t    `((*dealloc ,Dealloccount)
                (*JCall ,functionname)))))

% change branch to exitlabel into *exit when framesize ne 0 too

(fluid '(exitt&))

(de &jumpopt (u)
    (when (equal (cadar u) exitt&) (rplaca u '(*exit (*framesize)))))

% turn off lapopt in faslend

(fluid '(*lapopt))

(setq *lapopt t)

(when (funboundp 'old-faslend) (copyd 'old-faslend 'faslend))

(de faslend () 
    (prog (*lapopt)
       (return (old-faslend))))

% open codes for cons ncons xcons

(put 'cons 'opencode
      '((orh 16#4800 (reg 0) (reg 5))             % pair-tag
        (st!.l (reg 1) (displacement (reg r11) 0)) % Always sufficient memory for
        (st!.l (reg 2) (displacement (reg r11) 4))
        (or (reg r11) (reg 5) (reg 1)) % the result
        (*wplus2 (reg r11) 8)
        (*jumpwlessp (labelgen templabel) (reg r11) (reg r12)) 
        (*savefastframe)
        (*link !%reclaim expr 0)
        (*restorefastframe)  
        (labelref templabel))
)

(put 'xcons 'opencode
      '((orh 16#4800 (reg 0) (reg 5))             % pair-tag
        (st!.l (reg 2) (displacement (reg r11) 0)) % Always sufficient memory for
        (st!.l (reg 1) (displacement (reg r11) 4))
        (or (reg r11) (reg 5) (reg 1)) % the result
        (*wplus2 (reg r11) 8)
        (*jumpwlessp (labelgen templabel) (reg r11) (reg r12))
        (*savefastframe)
        (*link !%reclaim expr 0)
        (*restorefastframe)
        (labelref templabel))
)

(put 'ncons 'opencode
      '((orh 16#4800 (reg 0) (reg 5))             % pair-tag
        (st!.l (reg 1) (displacement (reg r11) 0)) % Always sufficient memory for
        (st!.l (reg nil) (displacement (reg r11) 4))
        (or (reg r11) (reg 5) (reg 1)) % the result
        (*wplus2 (reg r11) 8)
        (*jumpwlessp (labelgen templabel) (reg r11) (reg r12))
        (*savefastframe)
        (*link !%reclaim expr 0)
        (*restorefastframe)
        (labelref templabel))
)

(defcmacro *savefastframe)

(de  *savefastframe ()
   (cond ((null *usefastframe) '((*xalloc 1) (*move (reg 1) (xframe 1)) ))
         (T (append `((*xalloc ,(wplus2 nalloc* 1)) (*move (reg 1) (xframe 1)))
                     (pnth '((*move (reg ff4) (xframe 5))
                             (*move (reg ff3) (xframe 4))
                             (*move (reg ff2) (xframe 3)) 
                             (*move (reg ff1) (xframe 2))
                             (*move (reg nil) (reg nil))) % dummy for reduce
                         (wdifference 5 nalloc*))))))

(defcmacro *restorefastframe)

(de  *restorefastframe () 
   (cond ((null *usefastframe) '((*move (xframe 1) (reg 1))  (*xdealloc 1)))
        (T (append   (pnth '((*move (xframe 5) (reg ff4)) 
                             (*move (xframe 4) (reg ff3)) 
                             (*move (xframe 3) (reg ff2))  
                             (*move (xframe 2) (reg ff1))
                             (*move (reg nil) (reg nil))) % dummy for reduce 
                         (wdifference 5 nalloc*))
           `((*move (xframe 1) (reg 1)) (*xdealloc ,(wplus2 nalloc* 1))))) ))
 
(put 'bittable 'opencode
      '((shr 2 (reg 2) (reg 3)) % offset
        (ld!.b (indexed (reg 1) (reg 3)) (reg 5))
        (and 3 (reg 2) (reg 4))
        (shl 1 (reg 4) (reg 4))
        (subs 6 (reg 4) (reg 4))   % - fraction
        (shr  (reg 4) (reg 5) (reg 1))
        (and 3 (reg 1) (reg 1))))


(put 'putbittable 'opencode
      '((shr 2 (reg 2) (reg t1)) % offset
        (ld!.b (indexed (reg 1) (reg t1)) (reg 5))
        (and 3 (reg 2) (reg 4))   % %4
        (*move 3 (reg t2))
        (subs (reg t2) (reg 4) (reg 4))  % 3 - (pos%4)
        (shl 1 (reg 4) (reg 4))         % *2 / =fraction
        (shl  (reg 4) (reg t2) (reg t2))
        (andnot (reg t2) (reg 5) (reg 5))
        (and 3 (reg 3) (reg 3))        % value & 3
        (shl  (reg 4) (reg 3) (reg t3)) % <<fraction
        (or (reg 5) (reg t3) (reg 5))  % or
	(*wplus2 (reg 1) (reg t1))
        (st!.b (reg 5) (displacement (reg 1) 0))
        (*move (reg 3) (reg 1))))
 
% This code is intended for filling the 'holes' in the frame.
% e.g. in the following function
%
%  (de hugo (x) (cond ((null x) (call-another))
%                     (t (prog (a b c) ....))
%
% a frame of length 4 will be allocated, but it will not be preset (by nil)
% completely at entering the functiont but the frmae elements for a b and c
% will be set (to nil) if the prog is entered.
% So if the (null x) case is taken we have 3 unset frame elements on the stack
% which contain random information and this is dangerous for gc.
% to resolve that problem, the following function will work on the input
% to pass-1-lap and insert nil settings where needed.

(fluid '(&fillframeholes &frame-numbers))

(compiletime (flag '(findcmacro lookatinstruction) 'internalfunction)) 
 
(de &fillframeholes (code)
   (prog (allo cmactoextern externfound ll lll nn linearcode restcode)
      (when (not &fillframeholes) (return code))
      (setq cmactoextern
                  '(*link *linke *call *jcall *cons *xcons *fastmkfltn))
             % trivial tests whether we have to work at all
      (unless (setq allo (findcmacro code '*alloc)) (return code))
      (setq ll cmactoextern)
      (while (and (not externfound) ll)
            (setq externfound (findcmacro code (car ll))) 
            (setq ll (cdr ll)))  
      (when (not externfound) (return code))
 
            % so we have to work, starting at the *alloc 
 
      (setq nn (cadar allo))  %the frmae size              
      (setq linearcode t)
      (setq restcode (cdr allo))       
 
      (setq &frame-numbers nil) 
 
      (while (and linearcode restcode)
         (setq linearcode (lookatinstruction (car restcode)))
         (setq restcode (cdr restcode))) 
 
      (when linearcode (return code))
 
      (setq lll (setq ll (cons nil nil))) 
              
             % search frame that have not been set, prepare code,
             % put into place after *alloc
 
      (for (from i 1 nn 1) (do
            (unless (memq i &frame-numbers)
                  (rplaca ll `(*move (reg nil) (frame ,i)))
                  (when (eq &fillframeholes 'verbose) 
                    (prin2 "++ in ") (prin2 name&)
                    (prin2 " presetting frame ")
                    (prin2t i))
                  (push 'nil ll)))) 
      
      (when (eq lll ll) (return code))
      (rplacd lll (cdr allo))
 
      (rplacd allo (cdr ll))
      (return code))) 
         
(de lookatinstruction (instr)
      (cond ((atom instr) t)
            ((and (eq (car instr) '*move) 
                  (eqcar (caddr instr) 'frame)) 
                       (push (cadr (caddr instr)) &frame-numbers) 17)
            ((memq (car instr) '(*link *call *linke *jcall *jumpintype
                 *jumpeq *jump  *JumpType *JumpNotType *JumpNotInType
                 *jumpnoteq *jumpwlessp *jumpwgreaterp *jumpwleq *jumpwgeq
                 *jumpon)) nil)
            (t t)))
 
(de findcmacro (list pat) 
     (cond ((atom list ) nil) 
           ((eqcar (car list) pat) list) 
           (t        (findcmacro (cdr list) pat )))) 

(setq &fillframeholes t)
