% compiletime building table from compiler information

(compiletime 
  (progn (setq befs (mkvect 256))
         (de readbef (x)
             (when (get x 'opcodenumber)
                   (putv befs (land (get x 'opcodenumber) 255)
                        (cons x (get x 'opcodetype)))))
	 (dm genera(u) `(setq befvector ,befs))
         (mapobl (function readbef))))

(fluid '(befvector *using* *operands* *endofcode* *comments*))


(genera)

% stepping through the code

(de disassemble (fkt)
  (prog (where stop op fmt name)
    (when (idp fkt)
         (unless (fcodep fkt) (return nil))
         (setq where (symfnc (id2int fkt))))

    (setq *endofcode* nextbps) 
    (while (not stop)
       (setq *operands* nil *comments* nil)
       (printf "16#") (prinbn where 8)
       (printf "  ")
       (setq op (read-1-byte where))         
       (setq where (add1 where))
       (setq fmt (vecitm  (vecinf befvector) op))
       (when (atom fmt) (kaputt op))
       (setq name (car fmt))
       (setq fmt (cdr fmt))
       (cond ((eq fmt 'rr) (setq where (rr-format where name)))
	     ((eq fmt 'mx2)(setq where (mx2-format where)))
	     ((eq fmt 'mr1)(setq where (mr1-format where)))
             ((eq fmt 'rx) (setq where (rx-format where name)))
             ((eq fmt 'rs) (setq where (rs-format where)))
	     ((eq fmt 'rsm)(setq where (rsm-format where)))
	     ((eq fmt 'si) (setq where (si-format where)))
             ((eq fmt 'ss1)(setq where (ss-format where)))
	     ((eq fmt 'rs2)(setq where (rs2-format where)))
             ((eq fmt 'ss2)(setq where (ss-format where)))
             (t (kaputt fmt)))
       (tab 25) (prin2 name) (tab 35) (prin2 (car *operands*))
       (when *comments* (tab 60) (prin2 *comments*))
	(terpri)
       (when (greaterp where *endofcode*) (setq stop t))
    )
))

(de read-1-byte (where)
      (setq where (wand (byte where 0) 255))
      (prinbn where 2)
      where)

(de read-2-byte (where)
     (wor (wshift (read-1-byte where) 8)
          (read-1-byte (wplus2 where 1))))
      
(de rx-format(where name)
     (prog (ops op1 op2 op3 op4)
      (setq ops (read-1-byte where))
      (setq op1 (wshift ops -4))
      (setq op2 (wand ops 15))
      (setq ops (read-2-byte (wplus2 where 1)))
      (setq op3 (wshift ops -12))
      (setq op4 (wand ops 16#fff)) 
      (push (bldmsg "reg r%d, %d(reg r%d,reg r%d)" op1 op4 op2 op3)
                *operands*)

% try to find the end of the code (avoid disassemble of constants)

      (when (and *using* (or (eq 3 op2) (eq 3 op3)))
               (setq ops (wplus2 -1 (wplus2 *using* op4)))
               (when (wlessp ops *endofcode*)
                      (setq *endofcode* ops)))
  
      (return (wplus2 where 3))))

(de rs-format(where)
     (prog (ops op1 op2 op3 op4)
      (setq ops (read-1-byte where))
      (setq op1 (wshift ops -4))
      (setq op2 (wand ops 15))
      (setq ops (read-2-byte (wplus2 where 1)))
      (setq op3 (wshift ops -12))
      (setq op4 (wand ops 16#fff)) 
      (push (bldmsg "reg r%d, reg r%d, %d(reg r%d) " op1 op2 op4 op3)
                *operands*)
      (return (wplus2 where 3))))

(de rs2-format(where)
     (prog (ops op1 op2 op3 op4)
      (setq ops (read-1-byte where))
      (setq op1 (wshift ops -4))
      (setq op2 (wand ops 15))
      (setq ops (read-2-byte (wplus2 where 1)))
      (setq op3 (wshift ops -12))
      (setq op4 (wand ops 16#fff))
      (push (bldmsg "reg r%d, %d(reg r%d) " op1 op4 op3)
                *operands*)
      (return (wplus2 where 3))))


(de rsm-format(where)
     (prog (ops op1 op2 op3 op4)
      (setq ops (read-1-byte where))
      (setq op1 (wshift ops -4))
      (setq op2 (wand ops 15))
      (setq ops (read-2-byte (wplus2 where 1)))
      (setq op3 (wshift ops -12))
      (setq op4 (wand ops 16#fff))
      (push (bldmsg "reg r%d, %d, %d(reg r%d) " op1 op2 op4 op3)
                *operands*)
      (return (wplus2 where 3))))

(de ss-format(where)
     (prog (ops lng op1 op2 op3 op4)
      (setq lng (read-1-byte where))
      (setq ops (read-2-byte (wplus2 where 1)))
      (setq op1 (wshift ops -12))
      (setq op2 (wand ops 16#fff)) 
      (setq ops (read-2-byte (wplus2 where 3)))
      (setq op3 (wshift ops -12))
      (setq op4 (wand ops 16#fff)) 
      (push (bldmsg "%d(%d,reg r%d), %d(reg r%d) " op2 lng op1 op4 op3)
                *operands*)
      (return (wplus2 where 5))))

(de mx2-format(where)
     (prog (ops lng op1 op2 op3 op4)
      (setq lng (read-1-byte where))
      (setq op3 (wshift lng -4))
      (setq op4 (wand lng 15))
      (setq ops (read-2-byte (wplus2 where 1)))
      (setq op1 (wshift ops -12))
      (setq op2 (wand ops 16#fff))
      (push (bldmsg "%d, %d(reg r%d, reg r%d) " op3 op2 op4 op1)
                *operands*)
      (return (wplus2 where 3))))

(de si-format(where)
     (prog (ops lng op1 op2 op3 op4)
      (setq lng (read-1-byte where))
      (setq ops (read-2-byte (wplus2 where 1)))
      (setq op1 (wshift ops -12))
      (setq op2 (wand ops 16#fff))
      (push (bldmsg "%d, %d(reg r%d) " lng op2 op1)
                *operands*)
      (return (wplus2 where 3))))

(de mr1-format (where name)
     (prog (ops op1 op2)
      (setq ops (read-1-byte where))
      (setq op1 (wshift ops -4))
      (setq op2 (wand ops 15))
      (push (bldmsg "%d, reg r%d" op1 op2) *operands*)

      (return (wplus2 where 1))))

(de rr-format (where name)
     (prog (ops op1 op2)
      (setq ops (read-1-byte where))
      (setq op1 (wshift ops -4))
      (setq op2 (wand ops 15))
      (push (bldmsg "reg r%d, reg r%d" op1 op2) *operands*)

% try to find out "using" base point

      (when (and (eq op1 3) (eq op2 0) (eq name 'balr))
           (setq *using* (wplus2 where 1)))

      (return (wplus2 where 1))))

(de prinbn (it n)                % print an octal number
     (cond ((and (weq it 0) (wleq n 0)) nil)
            (t (progn
                  (prinbn (wshift it -4) (wplus2 n -1))
                  (prindig (wand it 15))
)    )     )  )

(de prindig (dig)        % print a numeric digit
     (if (wlessp dig 10) (writeChar  (wplus2 dig 48))
           (writechar (wplus2 65 (wdifference dig 10)))))

