%***********************************************************************
%
% (C) Copyright 1989
% File:             h-format.red
% Object File:      h-format
%
% Functions for formatting output of help files. To run the verification
% test of this program using the file h-test.txt enter the following:
%    htest "h-test.txt";
% The results explain themselves.
%
% Revision History: (Created 9/28/89)
%   11/4/89 Put the whole shebang into one file for easy compilation.
%   11/5/89 Fix @emph, @strong in h!/dpar function.
%   11/7/89 @table command added.
%   11/8/89 @noindent added.
%   11/12/89 Add @node, @eval, and @related commands.
%   11/23/89 Fix @itemize to copy starter each time, bind h!:enders.
%   11/1/91 Load the fast arith stuff at compile time only. Comment
%     some PC specific things that go in the h-sysdep.red file.
%  11/27/91 Move more PC specific stuff out.
%  12/17/91 Make sure things get interned for real St. Lisp.
%  12/18/91 Make verbatim check current character for real St. Lisp
%     systems (like CSL).
%  12/24/91 Fix what we believe EOF to be (not 26 always).
%
%***********************************************************************


% You must have h-rfast loaded or the .xxx operators defined.

global '(
  h!:tok                                 % Current token.
  h!:eof                                 % end of file token.
  h!:comchar                             % Command character (int).
  h!:lmar                                % Left margin.
  h!:rmar                                % Right margin.
  h!:pindent                             % Default indent.
  h!:tindent                             % Temporary indent.
  h!:enumlevel                           % @enumerate level.
  h!:curc                                % Current character.
  h!:eof                                 % End of file (token)
  h!:eol                                 % End of line.
  h!:tablecol                            % Default table column.
  h!:pos                                 % Current output column.
  h!:brkcom                              % Bracket commands.
  h!:syseof                              % System eof indicator.
        );

fluid '(
  h!:enders                              % List of ending tokens of para.
  h!:code!-offset                        % Code offset characters.
  h!:pskip                               % Paragraph skip vertical lines.
       );

h!:eof := '!$EOF!$;                      % End of file marker.
h!:syseof := 26;                         % End of file (system dependent).
h!:eol := '!$EOL!$;                      % End of line marker.
h!:comchar := 64;                        % Use @ as command character.
h!:lmar := 0;                            % Left margin default.
h!:rmar := 75;                           % Right margin default.
h!:pskip := 1;                           % Paragraph vertical skip (permanent).
h!:pindent := 1;                         % Permanent indent.
h!:tindent := 1;                         % Temporary indent.
h!:enumlevel := 0;                       % @enumerate level incremented when start.
h!:tablecol := 10;                       % @table second column.
h!:pos := 0;                             % Output column.

h!:code!-offset := '(32 124 32);         % Put a | in front of program.

% List defining the various commands that work on blocks of things.
h!:brkcom := 
           '(((64 101 109 112 104) . (h!/emph h!/deemph))
             ((64 115 116 114 111 110 103) . (h!/strong h!/weak))
             ((64 105 116) . (h!/it h!/norm))
             ((64 97 115 105 115) . (h!/asis h!/asis)) );


% Help system interface variables.
fluid '(
  NEXT!*                                 % Next topic (manual).
  PREVIOUS!*                             % Previous topic (manual).
  TOP!*                                  % Top level topic (manual).
  RELATED!*                              % Related topics (any order).
  RELATEDPTR!*                           % Pointer into this list.
       );


expr procedure htest f;
% The test module here. Errorset wrap for errors.
begin scalar fh;
  h!:curc := h!:tok := nil;
  fh := rds open(f, 'input);
  h!:lmar := 10; h!:rmar := 50;
  if atom errorset(quote h!/control(), t, t) then rds nil
    else close rds fh
end;


expr procedure h!/control;
% Read a file until done.
begin
  h!:tok := h!/ntok();
% prin2 ">> First token is "; print h!:tok;
loop: h!/comm t;
     if not(h!:tok eq h!:eof)  then go to loop
end;


expr procedure h!/comm ind;
% We're in paragraph mode. Read tokens until something signals end of
% paragraph, a command, or a blank line. If ind is T do the indent
% otherwise it's already done. Returns non-NIL if ok to continue, NIL if not.
begin scalar par, pare;
% prin2t ">> starting h!/comm";
  go to st;
loop1: h!:tok := h!/ntok();
% prin2 ">> loop1: tok is:"; print h!:tok;
st:  if h!:tok eq h!:eof then return nil
       else if h!:tok eq h!:eol then go to loop1
       else if eqcar(car h!:tok, h!:comchar) then
            if h!/linecomp() then return << h!/linecom(); t >>
              else if member(car h!:tok, h!:enders) then return nil
              else if h!/blockp() then return << h!/block(); t >>
              else if car h!:tok = '(64 98 121 101) then
                  return << h!:tok := h!:eof; t >>;
     par := pare := h!:tok;
% prin2t ">> starting paragraph";
loop2: h!:tok := h!/ntok();
     if h!:tok eq h!:eol then go to loop2s2;
loop2a: if h!:tok eq h!:eof then go to pend
      else if eqcar(car h!:tok, h!:comchar) then
          if h!/linecomp() then return << h!/dpar(ind, par); h!/linecom(); t >>
           else if member(car h!:tok, h!:enders) then
             return << h!/dpar(ind, par); t >>
           else if h!/blockp() then return << h!/dpar(ind, par); h!/block(); t >>
           else if car h!:tok = '(64 98 121 101) then return
             << h!:tok := h!:eof;  h!/dpar(ind, par); t >>;
     rplacd(lastpair pare, h!:tok);
     pare := lastpair pare;
% prin2 ">> Added another token:"; print pare;
     go to loop2;
loop2s2: h!:tok :=h!/ntok();
     if not(h!:tok eq h!:eol or h!:tok eq h!:eof) then go to loop2a;
pend: h!/dpar(ind, par);
  return t
end;

expr procedure h!/dpar(ind, l);
% Basic paragraph command. Only indent if ind is non-nil.
<< for i:=1:h!:pskip do terpri();           % Skip these lines.
   if ind then spaces (h!:pos := h!:lmar #+ h!:tindent);
   h!/dpar1 l;
   terpri() >>;

expr procedure h!/dpar1 l;
% l is a list of tokens. Put these into a paragraph in the current margins.
% Intercepts commands like @*, @br, @emph{...}, @strong{...}, and @it{...}.
begin scalar tk, ln, estack, anyextras, fspc, nspc;
   fspc := 0;  nspc := 1;
loop: if null l then
         << if estack then error(0, "missing } in this paragraph");
            h!:tindent := h!:pindent;
            return t >>;
     if car l = '(64 42) then            % Line break command.
         << terpri(); spaces(h!:pos := h!:lmar); fspc := 0; l := cdr l;
            go to loop >>
      else if car l = '(64 98 114) then  % Paragraph break command.
         << terpri(); terpri();
            spaces(h!:pos := h!:lmar #+ h!:pindent);
            fspc := 0; l := cdr l;
            go to loop >>
      else if tk := assoc(car l, h!:brkcom) then 
         << anyextras := eval list cadr tk;
            estack := list caddr tk . estack;
            if null(l := cdr l) then error(0, list(cadr tk, "missing {...}"));
            if car l neq '(123) then error(0, list(cadr tk, "missing {"));
            if null(l := cdr l) then error(0, list(cadr tk, "{ ... missing }"));
            if anyextras then << l := anyextras . l; nspc := 0 >>
                         else  go to loop >>
      else if car l = '(125) then
         << if null estack then error(0, "extra } in paragraph");
            anyextras := eval car estack;
            estack := cdr estack;
            l := cdr l;
            nspc := if l and memq(caar l, '(46 44 33 40 41 93 34 59 58 63 62))
                   then 0 else 1;
            if anyextras then << l := anyextras . l; fspc := 0 >>
                         else go to loop >>
      else nspc := 1;
     ln := length (tk := h!/ptoken car l) #+ fspc;
     if (h!:pos #+ ln) #> h!:rmar then
         << terpri(); spaces(h!:pos := h!:lmar); fspc := 0 >>;
     spaces fspc;
     fspc := nspc;
     h!:pos := h!:pos #+ ln;
     for each c in tk do writechar c;
     l := cdr l;
     go to loop
end;
     
expr procedure h!/length l;
% Computes length of a token. Ignore font change functions.
if null l then 0
  else if idp car l then h!/length cdr l
  else iadd1 h!/length cdr l;

expr procedure h!/punctp l;
% T if l is one of the punctuation commands.
 eqcar(l, 64) and cdr l and memq(cadr l, '(42 123 125 64));

expr procedure h!/ptoken l;
% l is a list of characters. If its a paragraph command, then do
% something about it. If its a punctuation @ then replace with appropriate
% characters.
 if h!/alphap l then h!/alpha l else h!/punct l;

expr procedure h!/punct l;
% Fix up @ punctuation in the word.
  if null l then nil
    else if eqcar(l, 64) then
       if cdr l then  cadr l . h!/punct cddr l
          else l
    else car l . h!/punct cdr l;

expr procedure h!/linecomp;
% Set h!:tok to command name if it's a line command.
begin scalar x;
  x := compressn cdar h!:tok;
  return if get(x, 'H!/LINE) then h!:tok := x
end;

expr procedure h!/linecom;
% u is a command token. Do the line command by reading all the tokens to
% eol and farm it out.
begin scalar cl, cle, cm;
  cm := h!:tok;     % Set above.
loop: if (h!:tok := h!/ntok()) eq h!:eof then return nil
       else if h!:tok neq h!:eol then
            << if cl then << rplacd(cle, h!:tok); cle := cdr cle >>
                 else cl := cle := h!:tok;
               go to loop >>;
  return apply(get(cm, 'H!/LINE), list cl)
end;

expr procedure h!/blockp;
% Set h!:tok to command name if it's a block command.
begin scalar x;
  x := compressn cdar h!:tok;
  return if get(x, 'H!/BLOCK) then h!:tok := x
end;

expr procedure h!/block;
% Farm out a block command. h!:tok is the function name to use.
  apply(get(h!:tok, 'H!/BLOCK), NIL);

expr procedure h!/noindent;
% Do the noindent command.
<< h!:tindent := 0;
   h!:tok := h!/ntok() >>;
put('noindent, 'H!/BLOCK, 'h!/noindent);


expr procedure h!/asis;
% Similar to IEFBR14.
  NIL;

%-----------------------------------------------------------------------
% Input scanner.

expr procedure h!/ntok;
% Scan the next input token: Skip blanks and tabs. Return <cr> as a
% special EOL token, and end of file as a special EOF token.
begin scalar tok, toke;
  if null h!:curc then h!:curc := readchar();
loop: if h!:curc eq h!:syseof then return h!:eof
        else if h!:curc eq 10 then return << h!:curc := readchar(); h!:eol >>
        else if h!:curc eq 64 then go to comm
        else if h!:curc #<= 32 then << h!:curc := readchar(); go to loop >>;
     tok := toke := list h!:curc;
     if h!:curc eq 123 or h!:curc eq 125 then 
       << h!:curc := readchar(); return list tok >>;
loop2: if (h!:curc := readchar()) #<= 32 or h!:curc eq 125 or h!:curc = 123
            then return list tok;
     rplacd(toke, list h!:curc);
     toke := cdr toke;
     go to loop2;
comm: h!:curc := readchar();
     tok := 64 . (toke := list h!:curc);
     if not((h!:curc #>= 65 and h!:curc #<= 90) or
            (h!:curc #>= 97 and h!:curc #<= 122)) then go to loop2;
loop3: h!:curc := readchar();
     if (h!:curc #>= 65 and h!:curc #<= 90) or
        (h!:curc #>= 97 and h!:curc #<= 122) then
       << rplacd(toke, list h!:curc);
          toke := cdr toke;
          go to loop3 >>;
     return list tok
end;

expr procedure h!/nextsep;
% Read next character and return T if separator.
  (h!:curc := readchar()) #<= 32 or h!:curc eq 123 or h!:curc eq 125;

expr procedure h!/scantoeol;
% Scan to the end of the input line and return the list of tokens.
begin scalar l, le;
loop: if (h!:tok := h!/ntok()) eq h!:eol then return l;
  if le then << rplacd(le, h!:tok); le := cdr le >>
     else l := le := h!:tok;
  go to loop
end;

%-----------------------------------------------------------------------
% Alphabetic type commands. These are replaced by a character or a token
% list. They occur inside paragraphs.

expr procedure h!/alphap l;
% T if l is an alphabetic command.
eqcar(l, 64) and cdr l and
  ((c #>= 65 and c #<= 90) or (c #>= 97 and c #<= 122) where c := cadr l);

global '(!*BREAK);

fluid '(l);
expr procedure h!/alpha l;
% Process the alphabetic command.  The characters are defined in the
% system dependent file.
begin scalar obr, x;
   obr := !*BREAK; !*BREAK := NIL;
   l := errorset('(compressn (cdr l)), nil, nil);
   !*BREAK := obr;
   if atom l then error(0, list(l, "invalid alphabetic command"))
     else l := car l;
   if x := get(l, 'H!/CHARNAME) then return list x
    else error(0, list(l, "invalid alphabetic command"))
end;
unfluid '(l);



%-----------------------------------------------------------------------
% Basic commands

expr procedure h!/center  l;
% Center this line between the margins. If this can't be done, just do
% it like a paragraph with no indent. Doesn't allow fancy things in the
% command line.
begin scalar sz, lw, ts;
  l := for each x in l collect h!/ptoken x;
  sz := (for each x in l sum length x) + sub1 length l;
  lw := iadd1(h!:rmar #- h!:lmar);
  if sz #> lw then h!/dpar(t, l)
    else << ts := h!:lmar #+ (lw #- sz) #/ 2;
            for each tk in l do
             << spaces ts; ts := 1;
                for each c in tk do writechar c >> >>;
  terpri();
end;
put('center, 'H!/LINE, 'h!/center);


expr procedure h!/verbatim ender;
% Copy the following characters "as-is" to the output stream. Write
% what ever is in h!:code-offset characters in front of each line.
% 'ender' is the list of characters that ends the verbatim mode.
begin scalar cl, c, ce;
  if h!:curc eq 10 then go to loop;
skp: if (c := readchar()) eq h!:eof then return h!:eof         
       else if c neq 10 then go to skp;
loop: if (c := readchar()) eq h!:eof then return h!:eof;
      if c = 13 then go to loop
        else if c neq 10 then
          << if null cl then ce := cl := list c
                 else << cdr ce := list c;  ce := cdr ce >>;
             go to loop >>;
    if cl neq ender then
      << spaces h!:lmar;
         for each c in h!:code!-offset do writechar c;
         for each c in cl do writechar c;
         terpri();
         cl := nil;
         go to loop >>;
  h!:tok := h!/ntok()
end;

expr procedure h!/example;
% This temporarily sets the offset to blanks. 
begin scalar h!:code!-offset;
  h!:code!-offset := '(32 32);
  h!/verbatim '(64 101 110 100 32 101 120 97 109 112 108 101)
end;
put('example, 'H!/BLOCK, 'h!/example);

expr procedure h!/program;
  h!/verbatim '(64 101 110 100 32 112 114 111 103 114 97 109);
put('program, 'H!/BLOCK, 'h!/program);

expr procedure h!/enumerate;
% Begin an enumerated list. Each item is signaled by @ITEM and the whole
% works ended with @END ENUMERATE. The levels available are:
%   arabic
%      roman numerals
%       alphabetic
begin scalar n, h!:enders;
  n := 1;
  h!:enders := '((64 105 116 101 109)     % @item
                 (64 101 110 100));       % @end 
  h!:lmar := h!:lmar #+ 3;
  h!:tok := h!/ntok();
  h!:enumlevel := iadd1 h!:enumlevel;
loop: if h!:tok eq h!:eol then h!:tok := h!/ntok()
      else if h!:tok eq h!:eof then 
        << h!:lmar := h!:lmar #- 3;
           h!:enumlevel := h!:enumlevel #- 1;
           return t >>
      else if atom h!:tok then error(0, h!:tok)
      else if car h!:tok = '(64 105 116 101 109) then     % @item
       << h!:tok := list nconc(h!/enum(h!:enumlevel, n), '(46));
          h!:tindent := -2;
          h!/comm t;
          n := n #+ 1 >>
      else if car h!:tok = '(64 101 110 100) then       % @end
       << h!:tok := h!/ntok();
          if car h!:tok neq '(101 110 117 109 101 114 97 116 101) then
            error(0, list("@end", compressn h!:tok, "should be @end enumerate"));
          h!:tok := h!/ntok();
          h!:lmar := h!:lmar #- 3;
          h!:enumlevel := h!:enumlevel #- 1;
          return t >>
      else if h!/linecomp() then h!/linecom()
      else if h!/block() then h!/block()
      else h!:tok := h!/ntok();
   go to loop
end;
put('enumerate, 'h!/block, 'h!/enumerate);


expr procedure h!/enum(lev, n);
if onep lev then explode2n n
  else if lev eq 2 then h!/roman n
    else if lev eq 3 then h!/alphabet n
    else h!/enum(remainder(lev, 3), n);


expr procedure h!/roman n;
% Generate roman numerals.
if izerop n then nil
  else if n > 100 then 99 . h!/roman(n #- 100)
  else if n >= 90 then 120 . 99 . h!/roman(n #- 90)
  else if n >= 50 then 108 . h!/roman(n #- 50)
  else if n >= 40 then 120 . 108 . h!/roman(n #- 40)
  else if n >= 10 then 120 . h!/roman(n #- 10)
  else if n = 9 then '(105 120)
  else if n >= 5 then 118 . h!/roman(n #- 5)
  else if n = 4 then '(105 118)
  else 105 . h!/roman(n #- 1);

expr procedure h!/alphabet n;
% Not very portable here but speed is everything.
if n > 26 then explode2n n else i2n(96 + n);


expr procedure h!/itemize;
% Read the next token. It should be some sort of single character to put
% in the front of each item. Then follow the enumerate code.
begin scalar ptk, h!:enders;
  ptk := h!/ntok();
  if h!/alphap ptk then ptk := h!/alpha ptk;
  h!:enders := '((64 105 116 101 109)     % @item
                 (64 101 110 100));       % @end 
  h!:lmar := h!:lmar #+ 3;
  h!:tok := h!/ntok();
loop: if h!:tok eq h!:eol then h!:tok := h!/ntok()
      else if h!:tok eq h!:eof then return(h!:lmar := h!:lmar #- 3)
      else if atom h!:tok then error(0, h!:tok)
      else if car h!:tok = '(64 105 116 101 109) then     % @item
       << h!:tok := append(ptk, nil);  % h!/comm uses RPLACD.
          h!:tindent := -2;
          h!/comm t >>
      else if car h!:tok = '(64 101 110 100) then       % @end
       << h!:tok := h!/ntok();
          if car h!:tok neq '(105 116 101 109 105 122 101) then
            error(0, list("@end", compressn h!:tok, "should be @end itemize"));
          h!:tok := h!/ntok();
          h!:lmar := h!:lmar #- 3;
          return t >>
      else if h!/linecomp() then h!/linecom()
      else if h!/block() then h!/block()
      else h!:tok := h!/ntok();
   go to loop
end;
put('itemize, 'h!/block, 'h!/itemize);


expr procedure h!/space n;
% Read the next token and skip that many blank lines.
<< n := compressn car n;
   for i:=1:n do terpri() >>;
put('sp, 'H!/LINE, 'h!/space);

%-----------------------------------------------------------------------
% Table command.

expr procedure h!/table;
% We use a default column of h!:tablecol
begin scalar comm, rolne, h!:enders, h!:pskip;
  comm := h!/ntok();
  if atom comm or not eqcar(car comm, 64) then
     error(0, "@table not followed by command (like @asis)");
  comm := append(comm, '((123)));
  if (h!:lmar := h!:lmar + h!:tablecol) >= h!:rmar - 10 then
     error(0, "margins too small for @table");
  h!:enders := '((64 101 110 100) (64 105 116 101 109)
                 (64 105 116 101 109 120));
  h!:pskip := 0;
loop: h!:tok := h!/ntok();
loop1: if h!:tok eq h!:eof then go done
        else if h!:tok eq h!:eol then go to loop
        else if h!:tok = '((64 101 110 100)) then
           if (h!:tok := h!/ntok()) = '((116 97 98 108 101)) then go done
             else error(0, "@table ... @end missing @end table")
        else if h!:tok = '((64 105 116 101 109)) then
             << if rolne then
                  << spaces(h!:pos := h!:lmar #- h!:tablecol);
                     h!/dpar1 rolne >>;
                rolne := append(comm, append(h!/scantoeol(), '((125)))) >>
        else if h!:tok = '((64 105 116 101 109 120)) then
             << if rolne then
                  << spaces(h!:pos := h!:lmar #- h!:tablecol);
                     h!/dpar1 rolne;
                     terpri() >>;
                rolne := append(comm, append(h!/scantoeol(), '((125)))) >>
        else << if rolne then
                  << spaces(h!:pos := h!:lmar #- h!:tablecol);
                     h!/dpar1 rolne >>;
                rolne := nil;
                if h!:pos >= h!:lmar then << terpri(); h!:pos := 0 >>;
                while h!:pos < h!:lmar
                      do << writechar 32;
                            h!:pos := iadd1 h!:pos >>;
                h!/comm NIL;
                go to loop1 >>;
     go to loop;
done: h!:lmar := h!:lmar #- h!:tablecol;
  terpri();
  h!:pos := 0;
  h!:tok := h!/ntok();
  return t
end;

put('table, 'h!/block, 'h!/table);

%-----------------------------------------------------------------------
% HELP system commands that are not removed by the preprocessor. These
% in the preprocessed format.

expr procedure h!/node;
% Read the next 3 numbers and put into NEXT!*, PREVIOUS!*, and TOP!*.
% Set h!:enders so that it terminates on next read.
<< NEXT!* := compressn car h!/ntok();
   PREVIOUS!* := compressn car h!/ntok();
   TOP!* := compressn car h!/ntok();
   h!:enders := '((64 110 111 100 101));
   h!:tok := h!/ntok() >>;
put('node, 'h!/block, 'h!/node);

expr procedure h!/reltd;
% Read tokens until the end of line and put into the RELATED!* list for
% later reference. Add each of these topics to the list of RELATED!* topics
% unless it is already there.
begin scalar l, le, x;
loop: if (h!:tok := h!/ntok()) eq h!:eol then 
           return << RELATED!* := nconc(RELATED!*, l);
                     if null RELATEDPTR!* then RELATEDPTR!* := l;
                     h!:tok := h!/ntok() >>;
    x := compressn car h!:tok;
    if member(x, RELATED!*) then go to loop;
    if le then << rplacd(le, list x); le := cdr le >>
     else l := le := list x;
    go to loop
end;
put('related, 'h!/block, 'h!/reltd);

expr procedure h!/eval;
% Read an expression and evaluate said. Any printing must be done inside the
% expression. We do a full read here to load readchar. We count parentheses,
% strings and the like so that the expression can go over one line.
begin scalar exp, l, le, pcnt, vcnt, anyc, obrk;
  l := le := list '! ;
  pcnt := vcnt := 0;
  go to loopa;
loop: h!:curc := readchar();
loopa: if h!:curc eq h!:syseof then error(0, "eof in @eval")
    else if h!:curc eq 33 then go to esc
    else if h!:curc eq 34 then go to str
    else if h!:curc eq 40 then anyc := pcnt := iadd1 pcnt
    else if h!:curc eq 41 then anyc := pcnt := isub1 pcnt
    else if h!:curc eq 91 then anyc := vcnt := iadd1 vcnt
    else if h!:curc eq 93 then anyc := vcnt := isub1 vcnt
    else if h!:curc #< 32 and not(h!:curc eq 13) and not(h!:curc eq 10)
          then go to loop;
  if h!:curc #> 32 then anyc := t;
  rplacd(le, list h!/raise h!:curc);
  le := cdr le;
  if not(anyc and izerop pcnt and izerop vcnt) then go to loop;
  obrk := !*BREAK; !*BREAK := NIL;
  errorset({'eval, {'internall, {'compress, mkquote l}}}, t, nil);
  h!:curc := readchar();
  h!:tok := h!/ntok();
  return (!*BREAK := obrk);
esc: rplacd(le, list h!/raise h!:curc);       % escaped character.
  le := cdr le;
  rplacd(le, list h!/raise readchar());
  le := cdr le;
  go to loop;
str: rplacd(le, list h!/raise h!:curc);       % String.
    le := cdr le;
    if (h!:curc := readchar()) eq h!:syseof then error(0, "eof in @eval")
      else if not(h!:curc eq 33) then go to str;
    rplacd(le, list '!");
    le := cdr le;
    if (h!:curc := readchar()) eq 33 then go to str else go to loop
end;
put('eval, 'h!/block, 'h!/eval);

expr procedure h!/raise c;
 n2i(if !*RAISE then if c #>= 97 and c #<= 122 then c #- 32 else c else c);

expr procedure internall x;
if idp x then intern x
  else if atom x then x
  else internall car x . internall cdr x;

end;


