%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PVNK:TRAP.SL
% Title:          Trap and interrupt handler for VAX PSL
% Author:         Eric Benson
% Created:        15 March 1982
% Modified:       17-Jul-84 14:15:38 (Brian Beach)
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983,  Hewlett-Packard Company, all rights reserved.
% Copyright (c) 1982 Eric Benson
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 11 Sep 1987 0731-MDT (Russ Fish)
%  Added catcher for sigpipe.
% 06 Dec 1984 1359-PST (Brian Beach)
%  Corrected order of arguments to SIGSET.  Restored NIL register
%  after SIGRELSE.
% 26 Jan 1984 1037-PST (Brian Beach)
%  Corrected file name in header.
% 2-Dec-83  16:00:00 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn
  (de *sigsetup (signame errorstring)
    % Signal handlers must re-enable after trapping
    `((PUSHAL (LIT (NOP)
		   (NOP)
		   (*MOVE (WCONST ,signame) (REG 1))
		   (*CALL EXTERNAL-SIGRELSE)
		   (*MOVE (FLUID NIL) (REG NIL))
		   (*MOVE (QUOTE ,errorstring) (REG 1))
		   (*JCALL STDERROR)
		   ))
      (*POP (REG 2))
      (*MOVE (WCONST ,signame) (REG 1))
      (*CALL EXTERNAL-SIGSET)
      ))

  (defcmacro *sigsetup)

  (define-constant  sighup 1)
  (define-constant  sigint 2)
  (define-constant  sigquit 3)
  (define-constant  sigill  4)
  (define-constant  sigtrap 5)
  (define-constant  sigiot  6)
  (define-constant  sigemt  7)
  (define-constant  sigfpe  8)
  (define-constant  sigkill 9)
  (define-constant  sigbus  10)
  (define-constant  sigsegv 11)
  (define-constant  sigsys  12)
  (define-constant  sigpipe 13)
  (define-constant  sigalrm 14)
  (define-constant  sigterm 15)
  ))

(lap '((*entry initializeinterrupts expr 0)
       (*sigsetup sigint  "Terminal interrupt")
       (*sigsetup sigill  "Illegal instruction")
       (*sigsetup sigtrap "Illegal instruction")
       (*sigsetup sigiot  "Illegal instruction")
       (*sigsetup sigemt  "Illegal instruction")
       (*sigsetup sigfpe  "Floating point exception")
       (*sigsetup sigbus  "Bus error")
       (*sigsetup sigsegv "Segmentation violation")
       (*sigsetup sigsys  "Bad argument to system call")
       (*sigsetup sigpipe  "write on a pipe with no one to read it")
       (*exit 0)))

