%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:              COS-EXTENSIONS
% Description:       Extended COS-Interface for PSL3.4
% Author:            W. Neun , ZIB Berlin
% Created:           10-Feb-86
% Mode:              Lisp
% Package:           NonKernel(Cray)
% Status:            Experimental (Do Not Distribute)
%
% (c) Copyright 1987, Konrad Zuse-Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(compiletime (load inum vector-instructions))
 
(compiletime (flag '(print-abort map-at1 whosdunnit pass-cos-args
                     dump-v-register fill-a-string) 'internalfunction))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Following LISP-functions allow the user to interface with certain
% useful COS-functions for Dataset-Management.
% The call corresponds with those FORTRAN-calls described in the
% COS-LIBRARY-REFERENCE-MANUAL with the exception of stat-parameter
% The stat will be the value of the Lisp-function, e.g.
%
% CFT : CALL ACQUIRE(STAT,'DN'L,'HUGO'L,'PDN'L,'meyer.b'L,'ED'L,'7'L,.)
%
% PSL : (cos-acquire '(DN HUGO PDN "meyer.b" ED "7" ..))
%                    returns stat.
%            Note : use string-quotes for lowercase of non-alpha chars
%
% or  : (Cos-Cmd "ACQUIRE,DN=HUGO,PDN='meyer.b',ED='7'.")
%
 
(Fluid '(*expected-interrupt *emsgp))
 
(setq *emsgp t)
 
(put 'acquire 'cos-functionnumber 0)
(put 'dispose 'cos-functionnumber 2)
(put 'fetch   'cos-functionnumber 4)
(put 'submit  'cos-functionnumber 6)
(put 'assign  'cos-functionnumber 8)
(put 'release 'cos-functionnumber 10)
(put 'access  'cos-functionnumber 12)
(put 'adjust  'cos-functionnumber 14)
(put 'delete  'cos-functionnumber 16)
(put 'modify  'cos-functionnumber 18)
(put 'save    'cos-functionnumber 20)
(put 'permit  'cos-functionnumber 22)
 
               % first three functions may cause interrupt
(de cos-acquire (l) (progn (setq *expected-interrupt t)
 (pass-cos-args l 'acquire) (setq *expected-interrupt nil)))
(de cos-dispose (l) (progn (setq *expected-interrupt t)
 (pass-cos-args l 'dispose) (setq *expected-interrupt nil)))
(de cos-fetch   (l) (progn (setq *expected-interrupt t)
 (pass-cos-args l 'fetch)   (setq *expected-interrupt nil)))
 
(de cos-submit  (l) (pass-cos-args l 'submit ))
(de cos-assign  (l) (pass-cos-args l 'assign ))
(de cos-release (l) (pass-cos-args l 'release))
(de cos-access  (l) (pass-cos-args l 'access ))
(de cos-adjust  (l) (pass-cos-args l 'adjust ))
(de cos-delete  (l) (pass-cos-args l 'delete ))
(de cos-modify  (l) (pass-cos-args l 'modify ))
(de cos-save    (l) (pass-cos-args l 'save   ))
(de cos-permit  (l) (pass-cos-args l 'permit ))
(de cos-rewind  (l)
       (when (and (pairp l) (pairp (cdr l)))
           (setq l (cadr l))
           (when (idp l) (setq l (symnam l)))
           (when (not (stringp l)) (return nil))
           (setq l (clear-rest-of-word l))
           (cos-backfile l)
           t))
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  COS command parser
%
%  format:
%
%   (COS-cmd  string)
%
%  string is a COS command in standard COS syntax enclosed by
%  PSL String quotes. Strings on parameter position have to be
%  enclosed by apostrophs. Unfortunately the name of the function
%  cannot be "COS" because this name is occupied by cosine.
%
%   example:
%   (COS-cmd "ACCESS,DN=RUMPEL,PDN='vectors.b',UQ.")
%  accessible are those COS commands, that are available from
%  PSL via a function call COS-xxx, e.g. COS-FETCH, COS-ACQUIRE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
(de COS-cmd (p)
   (prog (r w *raise)
    (when (not (stringp p))(conterror 4711
                                     "Cos Command: no string"
                                     (Cos-Cmd p)))
    (setq p (explode2 p))
mloop
    (setq *raise t)
    (cond ((null p) (go ready)))
    (setq w nil)
wloop
    (cond ((null p) (go word))
          ((memq (car p) '(!= !, !. ! )) (go word))
          ((eq (car p) '!')(go strng)))
    (setq w (cons (car p) w))
    (setq p (cdr p))
    (go wloop)
strng
    (setq *raise nil)
    (setq p (cdr p))
    (cond ((or (null p) (eqcar p '!')) (go strngend)))
    (setq w (cons '!! w))                 % escape for dot etc.
    (setq w (cons (car p) w))
    (go strng)
strngend 
    (when p (setq p (cdr p)))
    (when w (setq r (cons (compress (reverse w)) r)))
    (go mloop)
word
    (when p (setq p (cdr p)))
    (setq w (reverse w))        % numbers are converted to ids
    (when (and (wgeq (inf (car w)) (char !0))
               (wleq (inf (car w)) (char !9))) (setq w (cons '!! w)))
    (when w (setq r (cons (compress w) r)))
    (go mloop)
ready
    (setq r (reverse r))
    (setq r (cons
             (intern
              (compress
                 (append
                   '(C O S !-)
                    (explode2 (car r)) )))(cdr r)))
    (prin2t "***COS command function to be invoked:")
    (prin2t (list (car r) "'" (cdr r)))
    (return (idapply (car r) (list (cdr r)) ))
))
 
(Fluid '(cosargs arg1))
 
(De pass-cos-args (l name)
      (prog (li num elem vect)
        (setq li  l)
        (setq num 1)
        (setf (wgetv cosargs 1) (loc arg1))
        (setf arg1 0)
 
 loop
        (setq li (cdr li))
        (setq num (wplus2 num 1))
        (cond ((wgreaterp num 25) (prog2 (prin2t (list
                  "Too many parameters for cos-function" name))
                   (return nil))
        )     )
        (cond (li (go loop)))
 
        (setq vect (make-vector (wtimes2 num 30) 0))
        (setq vect (wplus2 (inf vect) 1))
        (setq num 2)
        (setq li l)
 loop1
        (setq elem (car li))
        (setq li (cdr li))
        (when (idp elem) (setq elem (symnam (idinf elem))))
        (if (intp elem) (setf (wgetv cosargs num) (inf li))
         (progn
         (when (not (stringp elem)) (nonstringerror elem name))
         (setf (wgetv cosargs num) vect)
         (clear-rest-of-word elem)
         (ifor (from i 0 (wrshift (wplus2 (strlen elem) 1) 3) 1)
               (do (setf (wgetv vect i) (wgetv elem (wplus2 i 1)))))
                   % copy string for COS 1.15
        ))
        (setq vect (wplus2 vect 30))
        (setq num (wplus2 num 1))
        (cond (li (go loop1)))
 
        (setf (wgetv cosargs 0) (wdifference num 1))
        (call-cos-function cosargs (get name 'cos-functionnumber))
        (return arg1)  % return stat value
)     )
 
% call-cos-function resides in the kernel (support)
 
(de clear-rest-of-word (strin)
    (prog (len len1)
         (setq len (add1 (strlen strin)))
         (setq len1 (plus (times 8 (quotient (plus len 8) 8)) -1) )
         (ifor (from i len len1 1)
               (do (putbyte (strbase (strinf strin)) i 0)
         )     )
         (return strin)
)   )
 
(de fill-a-string (strin stringlen)
%
% COS 1.16: a OWN and ID string must be filled to full size 15 chars
% with zeros
%
    (prog (len len1)
         (setq len (add1 stringlen))
         (ifor (from i len 15 1)
               (do (putbyte (strbase (strinf strin)) i 0)
         )     )
         (return strin)
)   )
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Set COS Default-ownership and id for Permanent Files %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Format : (cos-set-own "owner")
 
(de cos-set-own (strin)
   (if (stringp strin)
    (progn (cos-own   (fill-a-string (bldmsg "%w%w" strin "        ")
                             (strlen strin)))
           strin)
    (nonstringerror strin 'COS-Set-Own)
) )
 
% Format : (cos-set-id  "id")
 
(de cos-set-id (strin)
   (if (stringp strin)
    (progn (cos-id    (fill-a-string (bldmsg "%w%w" strin "        ")
                             (strlen strin)))
           strin)
    (nonstringerror strin 'COS-Set-Id)
) )
 
% LISP REPRIEVE processing. this routine is called from kernel (lisprpv)
% this function must be coded in LAP because it may not destroy Register
%
(Fluid '(undefnb00* *batch *cft-mode *dump *noreprieve *vregdump
        &numberofreprieves& ))
 
(setq   &numberofreprieves& 15)
 
(compiletime (off lapopt))
(compiletime (setq *syslisp nil))  % for NIL
(lap '((*entry lisp-reprieve expr 0)        % Lisp Reprieve-Routine
       (*move 5 (reg a3))                   %simulate an (alloc 3)
       (*Move (Reg B00) (Reg A6))           % save b00 and t00
       (*Move (Reg A6) (Reg 3))             % to print correct
       (*Move (Reg T00) (Reg 4))            % Register contents
       (*Move ($Fluid NIL) (Reg S6))      % restore nil
                                          % first rpv will be set
       (*Move (Reg S6) (Reg T00))         % by cosclrio
       (*Move ($Fluid *cft-mode) (Reg S6))
       (*Move (Reg S6) ($Fluid undefnb00*))   % saves an id
       (*jumpnoteq (label reload) (Reg S6) (quote nil))
     postreload
       (*jumpeq (label novregdump) ($fluid *vregdump) (quote nil))
       (*move 512 (reg a5))                % dumping vregs if requested
       (*move (reg Heaplast) (Reg a6))
       (*move (reg HeapTrapBound) (reg a7))
       (*wdifference (reg a7) (reg a5))
       (*jumpwgeq (label novregdump) (reg a6) (reg a7))
       (*move 64 (reg a5))
       (*move (reg a7) (reg s7))
       (*move (reg s7) ($fluid undefncode*))   % here   addr of dumparea
       (*vsetvl (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v0) (reg a0))
       (*wplus2 (reg a7) (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v1) (reg a0))
       (*wplus2 (reg a7) (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v2) (reg a0))
       (*wplus2 (reg a7) (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v3) (reg a0))
       (*wplus2 (reg a7) (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v4) (reg a0))
       (*wplus2 (reg a7) (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v5) (reg a0))
       (*wplus2 (reg a7) (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v6) (reg a0))
       (*wplus2 (reg a7) (reg a5))
       (*move (reg a7) (reg a0))
       (vstore (reg v7) (reg a0))
     novregdump
       (*jumpnoteq (label expect)
                  ($Fluid *expected-interrupt) (quote nil))
       (*move ($Fluid &numberofreprieves&) (reg s6))
       (*move 1 (reg s7))
       (*wdifference (reg s6) (reg s7))
       (*move (reg s6) ($Fluid &numberofreprieves&))
       (*move (reg s6) (reg s0))
       (jsp (label goon))
       (*move ($fluid *batch) (reg s6))
       (*move (quote nil) (reg s7))
       (*wxor3 (reg s0) (reg s6) (reg s7))
       (jsz (label goon))
       (*Move (quote "+++ Too many unexpected interrupts +++") (Reg 1))
       (*call console-print-string)
       (*Move (quote "+++ PSL aborts +++") (Reg 1))
       (*call console-print-string)
       (*call abort)
     goon
       (*Move -1 (Reg 1))        % tells no print at all
       (*jumpeq (label nodump) ($Fluid *emsgp) (quote nil))
       (*Move 0 (Reg 1))        % tells no dump, but a message
       (*jumpeq (label nodump) ($Fluid *dump) (quote nil))
       (*Move ($Fluid TRegSave) (Reg 1)) % tells to print a dump        O
       (*Move ($Fluid BRegSave) (Reg 2))
     nodump                                % set new rpv and eventually
       (*call dump-Registers expr 4)       % show Register-contents
       (*jumpnoteq (label norepr) ($Fluid *noreprieve) (quote nil))
       (*wand3 (Reg s6) (reg 1) 16#ffffff) %error address in loworder
       (*move (reg s6) (frame 1))
       (*wshift (Reg 1) -32)    % abcode is in the high-order bits
                                % after dump-Registers
       (*move (reg 1) (frame 2))
 
       (*jumpeq (label nemsg) ($Fluid *emsgp) (quote nil))
 
       (*jumpeq (label novregdump1) ($fluid *vregdump) (quote nil))
       (*move ($fluid undefncode*) (reg 1))   % here   addr of dumparea
       (*call dump-v-register)
     novregdump1
       (*move (frame 2) (reg 1))
       (*call print-abort)
       (*move (frame 1) (reg 1))
       (*call map-at)
       (*jumpnoteq (label cft) ($Fluid undefnb00*) (quote nil))
     try-to-continue
     nemsg
       (*move (frame 2) (reg 1))
       (*link user-error-function expr 1)
       (*move (frame 2) (reg 1))
       (*jumpnoteq (label notimelimit) (reg 1) 23) %dont continue
       (*call abort)
     notimelimit
       (*jumpnoteq (label nomemoryerr) (reg 1) 14) %dont continue
       (*call abort)
     nomemoryerr
       (*jumpeq (label nemsg2) ($Fluid *emsgp) (quote nil))
       (*Move (quote "*** PSL tries to continue ***") (Reg 1))
       (*call console-print-string)
     nemsg2
       (*Move (quote $error$) (Reg 1))
       (*Move (quote nil) (Reg 2))
       (*link throw expr 2)
     norepr
       (*Move (quote "*** Interrupt, full reprieve inhibited:") (Reg 1))
       (*call console-print-string)
       (*Move ($fluid *noreprieve) (Reg 1))
       (*call console-print-string)
       (*call abort)
     reload
       (*Move  ($Fluid BRegSave) (Reg A0))
       (*Move  64 (Reg A1))
       (BlkLod (Reg A1) (Reg B00))
       (*Move  ($Fluid TRegSave) (Reg A0))
       (BlkLod (Reg A1) (Reg T00))
       (*Move ($Fluid A3SAVE) (Reg A3))
       (*Move ($Fluid A4SAVE) (Reg A4))
       (*jump (label postreload))
     cft 
       (*Move (Reg T00) ($Fluid *cft-mode))     % *cft-mode = nil
       (*jumpeq (label try-to-continue) ($Fluid *batch) (quote nil))
       (*Move (quote "Interrupt in CFT-Code, no recovery when in batch")
            (Reg 1))
       (*call console-print-string)
       (*call abort)
     expect                         % some COS-functions will abort
       (*Move (Reg T00) ($Fluid *cft-mode))     % *cft-mode = nil
       (*Move (Reg T00) ($Fluid *expected-interrupt))
       (*Move -1 (Reg 1))                   % tells no printed dump
       (*link dump-Registers expr 5)        % set new reprieve
       (*jumpeq (label noemsg) ($Fluid *emsgp) (quote nil))
       (*Move (quote "*** An expected Interrupt has occured") (Reg 1))
       (*call console-print-string)
     noemsg
       (*Move (quote $error$) (Reg 1))
       (*Move (quote nil) (Reg 2))
       (*link throw expr 2)
)     )
(compiletime (setq *syslisp t))  % for NIL
(compiletime (on lapopt))
 
% user exit after reprieve-processing
% may be redefined
 
(de user-error-function (x) nil)
 
% print abort-codes  saves nextbps and options* and optionaddresses*
 
(Fluid '(options* nextbps *in-abort-print optionaddresses*
         cos-version*))
 
(setq cos-version* 114)   %default
 
(de print-abort (x)
  (if *in-abort-print nil
   (prog (myoptions* mynextbps myoptadd)
    (if (wgreaterp cos-version* 114)
      (progn
        (setq myoptions* (mkstring 50 32))
        (getmess x (strbase (strinf myoptions*)))
        (prin2t myoptions*))
      (progn
       (setq *in-abort-print t)
       (setq myoptions* options*)
       (setq myoptadd  optionaddresses*)
       (setq mynextbps nextbps)
       (load-to-psl 'print-abcode)
       (print-abcode x)
       (setq *in-abort-print nil)
       (setq options* myoptions*)
       (setq optionaddresses* myoptadd)
       (setq nextbps  mynextbps)
       (return x)
) ))))
 
(de map-at(x)
     (if heaptrapped (console-print-string
                        "+++ Error during Garbage Collection")
         (progn (prin2l (map-at1 x)) (terpri))))
 
(de map-at1 (x)    %looking for function causing the error
     (prog (min max minfkt maxfkt xx)
          (setq max nextbps)
          (setq min 0)
          (when (wgreaterp x nextbps)(return "*** outside code area"))
          (when (and (wgreaterp x symfnc)
                     (wlessp x (wplus2 symfnc maxsymbols)))
              (setq xx (wlshift (symfnc (wdifference x symfnc)) 2))
              (return (append  (whosdunnit xx optionaddresses*)
                      (list2
                      "; while entering" (mkid (wdifference x symfnc))))
          )   )
          (ifor (from y 0 maxsymbols 1)
              (do (cond ((funboundp (mkid y)) nil)   % those aren't of
                        ((flambdalinkp (mkid y)) nil) % any interest
                        ((not (stringp (symnam y))) nil) % unused id
                        ((and (wlessp (symfnc y) max)
                              (wgeq   (symfnc y) x  ))
                                 (setq max (symfnc y))
                                 (setq maxfkt (mkid y)))
                         ((and (wgeq   (symfnc y) min)
                               (wleq   (symfnc y) x  ))
                                 (setq min (symfnc y))
                                 (setq minfkt (mkid y)))
           )   )  )
    (return (append  (whosdunnit x optionaddresses*)
                     (list4 "; next external entries are" minfkt
                         "and" maxfkt))
)   ) )
 
(de whosdunnit (x optlist)
   (cond 
     ((null optlist) '("*** within code directly compiled to BPS"))
     ((and (wgreaterp x (cadar optlist))
              (wleq x (cddar optlist)))
                    (list2 "*** in load module " (caar optlist)))
     (t    (whosdunnit x (cdr optlist)))
) )
 
(de dump-v-register (addr)
  (prog (a b c)
   (console-newline)
   (console-print-string "*** Dump der Vector Register ***")
   (console-newline)
   (ifor (from vreg 0 7 1)
          (do (progn (terpri)
                (prin2 "Vectorregister ") (prin2t vreg)
                (ifor (from vregelt 0 62 3)
                       (do (progn
                            (setq a (getmem addr))
                            (setq b (getmem (wplus2 addr 1)))
                            (setq c (getmem (wplus2 addr 2)))
                            (setq addr (wplus2 addr 3))
                            (when (wlessp vregelt 10)
                             (terminalwritechar 32 32))
                            (putint vregelt)
                            (terminalwritechar 32 32)
                            (putoct a)
                            (terminalwritechar 32 32)
                            (putoct b)
                            (terminalwritechar 32 32)
                            (putoct c) (console-newline)
                )      )    )
                (setq a (getmem addr))
                (setq addr (wplus2 addr 1))
                (putint 63)
                (terminalwritechar 32 32)
                (putoct a)
                (console-newline)
 
   )      )    )
   (console-newline)
) )
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  the dm functions of "Library" CRAY PSL 3.4
%
(dm libopen (u)
 (if (cddr u)
  `(libopen1 ,(mkquote (cadr u)) ,(mkquote (caddr u)))
  `(libopen1 ,(mkquote (cadr u))  (quote input      ))))  % default
 
(dm libclose (u) `(libclose1 ,(mkquote (cadr u))))
 
(dm faslcopy (u)
      %  (faslcopy  mod1 mod2)   from mod1 to mod2 or
      %  (faslcopy  mod)         from mod  to mod
     (if (cddr u) `(faslcopy1 ,(cadr u) ,(caddr u))
                  `(faslcopy1 ,(cadr u) ,(cadr u) ) ))
 
