%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:EQUAL.SL
% Title:        EQUAL, EQN and friends
% Author:       Eric Benson
% Created:      19 August 1981
% Modified:     18-Jul-84 08:28:41 (Brian Beach)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 19 A, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 18-Jul-84 08:26:37 (Brian Beach)
%  Fixed bug in equal comparing bytes.
% 10-May-84 14:15:24 (Brian Beach)
%  <PSL.KERNEL>EQUAL.RED.3,  6-Feb-84 10:28:57, Edit by KESSLER
%  As per Hearn - EQUAL gives incorrect results for BIGNUMS - Changed
%  wordsequal and halfwordsequal.  Wasn't checking the last element.
%
% 22-Mar-84 11:22:01 (Brian Beach)
%  Fixed constants, tag names.
% 01-Dec-83 14:37:31 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 
(de eqn (u v)
  % Eq or numeric equality.
  (or (weq u v)
      (case (tag u)
      ((floatnum-tag)
       (and (floatp v)
            (weq (floathighorder u)
                 (floathighorder  v))))
      ((fixnum-tag)
       (and (fixnp v) (weq (fixval u) (fixval v))))
      ((bignum-tag)
       (and (bigp v) (wordsequal u v))) (nil nil))))
 
 
(de equal (u v)
  % Structural equality
  (or (weq u v)
      (case (tag u)
      ((evector-tag) (evectorequal u v))
      ((vector-tag)  (and (vectorp v) (vectorequal u v)))
      ((string-tag)
       (and (stringp v) (stringequal u v)))
      ((bytes-tag)
       (and (bytesp v) (stringequal u v)))
      ((pair-tag)
       (and (pairp v)
            (equal (car u) (car v))
            (equal (cdr u) (cdr v))))
      ((floatnum-tag)
       (and (floatp v)
            (weq (floathighorder u)
                 (floathighorder v))))
      ((fixnum-tag)
       (and (fixnp v)
            (weq (fixval  u) (fixval  v))))
      ((bignum-tag) (and (bigp v) (wordsequal u v)))
      ((words-tag) (and (wrdsp v) (wordsequal u v)))
      ((halfwords-tag) (and (halfwordsp v) (halfwordsequal u v)))
      (nil nil))))
 
(de eqstr (u v)
  % Eq or string equality.
  (or (weq u v)
      (and (stringp u) (stringp v) (stringequal u v))))
 
(de stringequal (u v) (unchecked-string-equal u v))
 
(de wordsequal (u v)
  (prog (s1 i)
        (setq u (wrdinf u))
        (setq v (wrdinf v))
        (unless (weq (setq s1 (wrdlen u)) (wrdlen v))
          (return nil))
        (setq i 0)
   loop
        (when (wgreaterp i s1)
          (return t))
        (unless (weq (wrditm u i) (wrditm v i))
          (return nil))
        (setq i (wplus2 i 1))
        (go loop)))
 
(de halfwordsequal (u v)
  (prog (s1 i)
        (setq u (halfwordinf u))
        (setq v (halfwordinf v))
        (unless (wgreaterp (setq s1 (halfwordlen u)) (halfwordlen v))
          (return nil))
        (setq i 0)
   loop
        (when (wgreaterp i s1)
          (return t))
        (unless (weq (halfworditm u i) (halfworditm v i))
          (return nil))
        (setq i (wplus2 i 1))
        (go loop)))
 
(de vectorequal (u v)
  % Vector equality without type check
  (prog (len i)
        (setq u (vecinf u))
        (setq v (vecinf v))
        (setq len (veclen u))
        (when (wneq len (veclen v))
          (return nil))
        (setq i 0)
   loop
        (when (wgreaterp i len)
          (return t))
        (unless (equal (vecitm u i) (vecitm v i))
          (return nil))
        (setq i (wplus2 i 1))
        (go loop)))
 
(de evectorequal (u v)
  % E-Vector equality without type check
  (prog (handler)
        (when (and (not (funboundp 'object-get-handler-quietly))
                   (setq handler (object-get-handler-quietly u 'equal)))
          (return (apply handler (list u v))))))
