 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 % PT:PRINTF.RED - Formatted print routine
 %
 % Author:      Eric Benson, M. Griss
 %             Symbolic Computation Group
 %              Computer Science Dept.
 %              University of Utah
 % Date:        27 August 1981
 % Copyright (c) 1981 University of Utah
 %
 % Revisions:
 % 5/9/85 - Kessler - Sleaze of the year.  Changed printf1 to not
 %  Do any pushes
 %  <PSL.NEWTEST>PRINTF.RED.9, 11-Aug-83 11:43:32, Edit by KESSLER
 %  Added ALLOC 0 to printf1.
 % MLGriss, 17 July 1983
 %       Removed FLUID binding of FORMATFORPRINTF* in PRINTF, PRINTF1
 %      Remove FLUID binding of OUT* in ChannelPrintF
 %      Add some common SMACROS
 %  <PSL.KERNEL>PRINTF.RED.2, 17-Sep-82 16:01:01, Edit by BENSON
 %  Added ChannelPrintF
 %  <PSL.INTERP>PRINTF.RED.6,  3-May-82 10:45:11, Edit by BENSON
 %  %L prints nothing for NIL
 %  <PSL.INTERP>PRINTF.RED.9, 23-Feb-82 21:40:31, Edit by BENSON
 %  Added %x for hex
 %  <PSL.INTERP>PRINTF.RED.7,  1-Dec-81 16:11:11, Edit by BENSON
 %  Changed to cause error for unknown character
 
(COMPILETIME (FLAG '(PRINTF1 PRINTF2) 'INTERNALFUNCTION))
 
(compiletime (load sys-consts token-decls))
 
 
(FLUID '(FORMATFORPRINTF* lineposition tokenbuffer))
 
 
 % First, set FormatForPrintF*
 
(DE PRINTF (FMT A1 A2 A3 A4 A5 A6 A7 A8 A9 A10 A11 A12 A13 A14)
  (PROGN (SETQ FORMATFORPRINTF* FMT)
         (PRINTF1 FMT A1 A2 A3 A4 A5 A6 A7 A8 A9 A10 A11 A12 A13 A14)))
 
 
 
 % Then, push all the registers on the stack and set up a pointer to them
 
(LAP '((*ENTRY PRINTF1 EXPR 15) (*ALLOC 14) (*MOVE (REG 2) (FRAME 1))
       (*MOVE (REG ST) (REG 1)) (*WPLUS2 (REG 1) (WCONST 2))
       (*MOVE (REG 3) (FRAME 2)) (*MOVE (REG 4) (FRAME 3))
       (*MOVE (REG 5) (FRAME 4)) (*MOVE (REG 6) (FRAME 5))
       (*MOVE (REG 7) (FRAME 6)) (*MOVE (REG 8) (FRAME 7))
       (*MOVE (REG 9) (FRAME 8)) (*MOVE (REG 10) (FRAME 9))
       (*MOVE (REG 11) (FRAME 10)) (*MOVE (REG 12) (FRAME 11))
       (*MOVE (REG 13) (FRAME 12)) (*MOVE (REG 14) (FRAME 13))
       (*MOVE (REG 15) (FRAME 14)) (*CALL PRINTF2) (*EXIT 14)) )
 
 
 
 % Finally, actual printf, with 1 argument, pointer to array of parameters
 % written on the currently selected output channel.  The exception to this is
 % that when a % is encountered, the following character is interpreted as a
 % format character, to decide how to print one of the other arguments.  The
 % following format characters are currently supported:
 %      %b - blanks; take the next argument as integer and print that many
 %            blanks
 %      %c - print the next argument as a single character
 %      %d - print the next argument as a decimal integer
 %       %e - EVALs the next argument for side-effect -- most useful if the
 %            thing EVALed does some printing
 %      %f - fresh-line, print end-of-line char if not at beginning of line
 %      %l - same as %w, except lists are printed without top level parens
 %      %n - print end-of-line character
 %      %o - print the next argument as an octal integer
 %      %p - print the next argument as a Lisp item, using Prin1
 %       %r - print the next argument as a Lisp item, using ErrPrin (`FOO')
 %      %s - print the next argument as a string
 %      %t - tab; take the next argument as an integer and
 %            print spaces to that column
 %      %w - print the next argument as a Lisp item, using Prin2
 %      %x - print the next argument as a hexidecimal integer
 %      %% - print a %
 %
 % If the character is not one of these (either upper or lower case), an
 % error occurs.
 %
(DE PRINTF2 (PRINTFARGS)
  (PROG (UPLIM I CH UPCH FMT )
        (SETQ FMT FORMATFORPRINTF*)
        (SETQ UPLIM (STRLEN (STRINF FMT)))
        (SETQ I 0)
        (WHILE (WLEQ I UPLIM)
               (PROGN (SETQ CH (STRBYT (STRINF FMT) I))
                      (COND ((WNEQ CH (CHAR !%)) (WRITECHAR CH))
                            (T (PROG NIL
       (SETQ I (WPLUS2 I 1))
       (SETQ CH (STRBYT (STRINF FMT) I))
       (SETQ UPCH
        (COND ((LOWERCASECHAR CH)
               (RAISECHAR CH))
              (T CH)))
       (CASE UPCH
        (((CHAR !B))
         (PROGN (SPACES (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !C))
         (PROGN (WRITECHAR
                 (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !D))
         (PROGN (WRITESYSINTEGER
                 (GETMEM PRINTFARGS) 10)
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !E))
         (PROGN (EVAL (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !F))
         (COND ((WGREATERP (POSN) 0)
                (WRITECHAR (CHAR EOL)))))
        (((CHAR !L))
         (PROGN (PRIN2L (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !N)) (WRITECHAR (CHAR EOL)))
        (((CHAR !O))
         (PROGN (WRITESYSINTEGER
                 (GETMEM PRINTFARGS) 8)
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !X))
         (PROGN (WRITESYSINTEGER
                 (GETMEM PRINTFARGS) 16)
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !P))
         (PROGN (PRIN1 (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !R))
         (PROGN (ERRPRIN (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !S))
         (PROGN (WRITESTRING
                 (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !T))
         (PROGN (TAB (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !W))
         (PROGN (PRIN2 (GETMEM PRINTFARGS))
                (SETQ PRINTFARGS
                 (LOC
                  (WGETV PRINTFARGS
                   STACKDIRECTION)))))
        (((CHAR !%)) (WRITECHAR (CHAR !%)))
        (NIL
         (STDERROR
          (BLDMSG
           '"Unknown character code for PrintF: %r"
           (MKID CH))))))))
                      (SETQ I (WPLUS2 I 1))))))
 
 
      % also A5..A14
(DE ERRORPRINTF (FORMAT A1 A2 A3 A4)
  (PROG (SAVECHANNEL)
        (SETQ SAVECHANNEL (WRS ERROUT*))
        (COND ((WGREATERP (WGETV LINEPOSITION (INTINF  ERROUT*))
                          0)
               (TERPRI)))
        (PRINTF FORMAT A1 A2 A3 A4)
        (COND ((WGREATERP (WGETV LINEPOSITION (INTINF  ERROUT*))
                          0)
               (TERPRI)))
        (WRS SAVECHANNEL)))
 
 
    % shares TokenBuffer
(DE TOSTRINGWRITECHAR (CHANNEL CH)
  (PROGN (COND ((WGEQ (WGETV TOKENBUFFER 0) (WDIFFERENCE MAXTOKENSIZE 1))
                (PROGN (SETF (WGETV TOKENBUFFER 0) 80)
                       (SETF (STRBYT TOKENBUFFER 80) (CHAR NULL))
                       (STDERROR
          (LIST '"Buffer overflow while constructing error message:"
                  Formatforprintf*
         '"The truncated result was:"
         (COPYSTRING (MKSTR TOKENBUFFER))))))
               (T (PROGN (SETF (WGETV TOKENBUFFER 0)
 (WPLUS2 (WGETV TOKENBUFFER 0) 1))
                         (SETF (STRBYT TOKENBUFFER (WGETV TOKENBUFFER 0))
 CH))))))
 
 
     %. Print to string
     % takes up to 14 args
(DE BLDMSG (FORMAT AR1 AR2 AR3 AR4)
  (PROG (TEMPCHANNEL OUT*)
        (SETF (WGETV LINEPOSITION 2) 0)
        (SETF (WGETV TOKENBUFFER 0) -1)
        (SETQ TEMPCHANNEL OUT*)
        (SETF OUT* '2)
        (PRINTF FORMAT AR1 AR2 AR3 AR4)
        (SETF (STRBYT (STRINF TOKENBUFFER)
                      (WPLUS2 (WGETV TOKENBUFFER 0) 1))
              (CHAR NULL))
        (SETF  OUT* TEMPCHANNEL)
        (RETURN (COPYSTRING TOKENBUFFER))))
 
 
     %. `Prin1 with quotes'
(DE ERRPRIN (U)
  (PROGN (WRITECHAR (CHAR !`))
         (PRIN1 U)
         (WRITECHAR (CHAR !'))))
 
 
 
    %. Prin2 without top-level parens
    % NIL is (), print nothing
(DE PRIN2L (ITM)
  (COND ((NULL ITM) NIL)
        ((NOT (PAIRP ITM)) (PRIN2 ITM))
        (T (PROGN (WHILE (PROGN (PRIN2 (CAR ITM))
  (SETQ ITM (CDR ITM))
  (PAIRP ITM))
                         (CHANNELWRITEBLANKOREOL OUT*))
                  (COND (ITM (PROGN (CHANNELWRITEBLANKOREOL OUT*)
      (PRIN2 ITM))))))))
 
(DE CHANNELPRINTF (CHN FORMAT A1 A2 A3 A4 A5 A6 A7 A8 A9 A10 A11 A12 A13)
  (PROG (SAVECHN)
        (SETQ SAVECHN OUT*)
        (SETQ OUT* CHN)
        (PRINTF FORMAT A1 A2 A3 A4 A5 A6 A7 A8 A9 A10 A11 A12 A13)
        (SETQ OUT* SAVECHN)))
