%
% aegis_load.sl - Invoke the Aegis loader for C/Pascal/etc routines.
%
% Author:      John W. Peterson
%              Computer Science Dept.
%              University of Utah
% Date:        June 1984
%
% Important: See notes near the bottom if you plan on seriously using this.
%
(compiletime (load useful))   % uses for loops
% The following is lifted out of ~s/us/ins/loader.ins.pas

%
%    PM_$Opts =                                                          
%        (PM_$Copy_Proc,             { copy procedure text }             
%         PM_$Install,               { install defined globals }         
%         PM_$No_Unresolveds,        { don't allow unresolved globals }  
%         PM_$Load_Global,           { load in global space }            
%         PM_$Install_Sections,      { install defined sections }        
%         PM_$Load_Writable) ;       { load the object module writable } 
%                                                                        
%    PM_$Loader_Opts = Set of PM_$Opts;                                  
%
%    PM_$Sect_Info = Record
%        Name:       KG_$Name_String;
%        Loc:        Linteger;
%        Len:        Linteger
%        End;
%
%    PM_$Load_Info = Record                                          
%        Start_Addr: Univ_Ptr;       { Start address (nil if none) } 
%        N_Sects:    Integer;        { Number of Sections }          
%        Sects:      Array[1..1] of PM_$Sect_Info                    
%        End;                                                        
%                                                                    
%
(de PM_Load (filename opts N_sects)
  (prog(namelen loadinfoptr status
        start-addr
        Nsects
        sects)
    (setf namelen (StringLength filename))
    (putarg 1 Filename '(string))
    (putarg 2 namelen '(integer16))
    (putarg 3 opts '(integer16))
    (putarg 4 N_sects '(integer16))
    (setf loadinfoptr (gtWrds 12))
    (putObjPointer 5 loadinfoptr)
    (clearArg 6)
    (syscall 'PM_!$LOAD 6)

    (setf status (getarg 6 '(integer32) '(FixN)))
    (return
      (cond ((eq status 0)
             (ApolloResetStatus)
             % The usefulness of these is questionable.
             (setf start_addr (getword loadinfoptr 0))
             (setf Nsects (gethalfword loadinfoptr 2))
             (list start_addr Nsects) )
             % ignore tearing the rest of it apart for now
  
            (T (ApolloError status))
      )
    )
  )
)

%
% This call is like doing a shell "inlib", but it can be done
% from a running lisp. 
% The filename is assumed to be an Aegis binary.  Once loaded, SysCalls
% can access the functions just like any other global routine.
%
(de AegisLoad(filename)
  (pm_load filename 3 0)  % "3" options just seem to work...
)

% Force Aegis to look up an entry, used if a .bin file is reloaded.

(de AegisRedefine(routine-name)
  (cond ((PairP routine-name)   % Have a list of them?
         (for (in routine routine-name)
              (do (remprop routine 'AegisEntryAddress))))

        (T                      % Assume single ID
          (remprop routine-name 'AegisEntryAddress))
  )
)

% NOTE!  If you re-load the same aegis binary file, you must execute
% AegisRedefine for each procedure defined in the binary file.
% This becuase SysCalls "cleverly" stashes the physical entry
% address to avoid looking it up each time.
%
% Make the following mods to ~p/sup/syscalls.red if you're going to be debugging
% things:
%   1) Add !*ForceAegisLookup to the list of fluids
%   2) Add 
%         !*ForceAegisLookup := NIL; 
%      to the statements executed at LoadTime
%
%   3) Change LookUpAegisEntry to look like this:
%
%        Symbolic Procedure LookUpAegisEntry(EntryPointName);
%        begin scalar AddressOfAegisRoutine;
%         if IDP EntryPointName
%          then if (not !*ForceAegisLookup) and
%                  (AddressOfAegisRoutine:=get(EntryPointName,'AegisEntryAddress))
%                then return AddressOfAegisRoutine
%                      :
%                      :    % (rest of routine stays the same)
%        
% Set the *ForceAegisLookup flag to T if you're in a debugging loop.  This 
% causes SysCalls to always look up the value from the OS, but slows things
% down considerably.


