%
% Graphics Primatives calls for Apollo PSL.  (first file).
%
% Author:      John W. Peterson
%              Computer Science Dept.
%              University of Utah
% Date:        December 1982
%
%  Based on:
%  GPR.INS.PAS, /sys/ins, dpg, 04/01/82
%  graphics primitive package definitions
%  by Apollo Comp. Co.
%
% For further documentation on the GPR package see chapter 12 of the Apollo
%   System Programmer's Guide.
%
% 25-JUN-83, Edit by JW Peterson, merge GPR1_syscalls and GPR2_syscalls, also
%  add some fixes (e.g., those for color node).
% 28-Jun-83, Edit by JWP, add some Direct mode calls & fix clipping stuff.
%  Initialization and termination.
%  4-Jan-84, Edit by JW Peterson, added some basic polygon fill stuff.
% 13-Mar-84, JWP, Got rid of some spurious fluids
% 13-mar-84, Edit by JW Peterson, fixed handling of returned status.
% 20-may-84, Edit by JW Peterson, added some fixes sent in
%               by Susan Ennis of Amoco (haven't tested them)
% 27-jun-84, Edit by JW Peterson, fixed gpr_set_cursor_origin to use
%               PutObjPointer

%  GPR_INIT initializes the graphics primitive package.

(de gpr_init 
                (op                     % mode of operation
                unit_or_pad             % display unit or stream id of dm pad if any
                size                    % disp bitmap sizes in bits for x and y
                hi_plane)               % highest plane number disp bitmap is to have

            

(prog (status bitmapsize 
             init_bitmap)               % descriptor of initial bitmap;
  (setq bitmapsize (gtwrds 1))          % Allocate space for Lisp Arg
  (puthalfword bitmapsize 0 (first size)) 	% Retrofit Lisp dotted pair into Pascal
  (puthalfword bitmapsize 1 (second size)) 	% record.

  (putarg 1 op '(integer16)) 
  (putarg 2 unit_or_pad '(integer16)) 
  (putobjpointer 3 bitmapsize) 
  (putarg 4 hi_plane '(integer16)) 

  (cleararg 5) 
  (cleararg 6) (syscall 

  'gpr_$init 6) 

  (setq status (getarg 6 '(integer32) '(fixn))) 

  (cond ((eq status 0) (progn 
    (apolloresetstatus) 
    (setq init_bitmap (getarg 5 '(integer32) '(fixn))) 
    (return init_bitmap) nil)) (t 
  
  (return (apolloerror status))))))


%  GPR_TERMINATE terminates this package's operation.

(de gpr_terminate 
                (delete_disp)           % whether to delete dm frame
            
(prog (status) 
  (putarg 1 delete_disp '(integer16)) 
  (cleararg 2) (syscall 

  'gpr_$terminate 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


%  Set and inquire operations for the display.
%  GPR_INQ_CONFIG returns the current display configuration.

(de gpr_inq_config nil 

(prog (config status) 
  (cleararg 1) 
  (cleararg 2) (syscall 
  'gpr_$inq_config 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 
  (cond ((eq status 0) (progn 
    (apolloresetstatus) 
    (setq config (getarg 1 '(integer16) '(inum))) 
    (return config))) (t 
  
    (return (apolloerror status))))))


%  GPR_SET_COLOR_MAP gives new values for the color map.

(fluid '(gpr_valvect))
  (setq gpr_valvect (gtbps 255))

(de gpr_set_color_map (start vallist) 
(prog (index mixcolor status) 

  (putarg 1 start '(integer32)) 
  (putarg 2 (length vallist) '(integer16)) 
  (setq index 0) 
  (foreach i in vallist do (progn 
    (setq mixcolor (plus (times (indx i 0) 65536) (times (indx i 1) 256) (indx i 2))) 
    (putword gpr_valvect index mixcolor) 
    (setq index (plus index 1)) nil)) 
  

  (putobjpointer 3 gpr_valvect) 
  (cleararg 4) (syscall 

  'gpr_$set_color_map 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



%  GPR_INQ_COLOR_MAP returns current values in the color map.

%PROCEDURE   gpr_inq_color_map (
%                start_index,  % index of first entry
%                n_entries,    % # of entries
%begin Scalar v, Status;
%  NOT YET IMPLEMENTED, wait 'til we get a color node.
%end;

%  GPR_SET_CURSOR_PATTERN loads a cursor pattern.

(de gpr_set_cursor_pattern 
               (cursor)                 % pattern for cursor
	       
(prog (status) 
   (putarg 1 cursor '(integer32)) 
   (cleararg 2) (syscall 

   'gpr_$set_cursor_pattern 2) 

   (setq status (getarg 2 '(integer32) '(fixn))) 

   (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


%  GPR_SET_CURSOR_ACTIVE specifies whether the cursor should be on or off.

(de gpr_set_cursor_active 
                (active)                % false --> off, true --> on
            
(prog (status) 
   (putarg 1 status '(integer16)) 
   (cleararg 2) (syscall 

   'gpr_$set_cursor_active 2) 

   (setq status (getarg 2 '(integer32) '(fixn))) 
 
   (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


%  GPR_SET_CURSOR_POSITION gives the position at which the cursor is
%   to be displayed, in the current bitmap.

(de gpr_set_cursor_position 
                (pos)                   % where it goes
	    
(prog (status cursorpos) 
  (setq cursorpos (gtwrds 1))           % Allocate space for Lisp Arg
  (puthalfword cursorpos 0 (first pos)) % Retrofit Lisp dotted pair into Pascal
  (puthalfword cursorpos 1 (second pos)) 	% record.

  (putarg 1 cursorpos '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$set_cursor_position 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



%  GPR_SET_CURSOR_ORIGIN gives the cursor-relative position of its pixel
%   which is to be placed at the cursor position.

(de gpr_set_cursor_origin 
                (origin)                % the position of the cursor pixel
            

(prog (status cursororg) 
  (setq cursororg (gtwrds 1))           % Allocate space for Lisp Arg
  (puthalfword cursororg 0 (first origin)) 	%Retrofit Lisp dotted pair into Pascal
  (puthalfword cursororg 1 (second origin)) 	% record.

  (putobjpointer 1 cursororg '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$set_cursor_origin 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



%  GPR_INQ_CURSOR returns information about the cursor.

(de gpr_inq_cursor nil 
(prog 
         (cursor                        % pattern for cursor
         ops                            % raster ops for cursor
         active                         % whether cursor is active
         pos                            % current cursor position
         origin                         % current cursor origin
         status)                        % returned status

  (cleararg 1) 
  (cleararg 2) 
  (cleararg 3) 
  (setq pos (gtwrds 1))                 %Allocate memory for return argument
  (putobjpointer 4 pos)                 %Place pointer to it in stack fram to accept...
				%..returned data
  (setq origin (gtwrds 1)) 
  (putobjpointer 5 origin) 
  (cleararg 6) (syscall 

  'gpr_$inq_cursor 6) 

  (setq status (getarg 6 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn (apolloresetstatus) 
            (setq cursor (getarg 1 '(integer32) '(fixn))) 
            % HalfWord array?????            ops
            (setq active (getarg 3 '(integer16) '(inum))) 
            (setq pos (cons (gethalfword pos 0) (gethalfword pos 1))) 
            (setq origin (cons (gethalfword origin 0) 
	              (gethalfword origin 1))) 
	    (return (list pos active origin)) nil)) (t 
	 (return (apolloerror status))))))

	
%eject;
%  Bitmap control functions.



%  GPR_SET_BITMAP establishes the current bitmap for subsequent operations.

(de gpr_set_bitmap 
                (bitmap)                %  the given bitmap
            
(prog (status)                          % returned status
  (putarg 1 bitmap '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$set_bitmap 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


%  GPR_INQ_BITMAP returns a descriptor of the current bitmap.

(de gpr_inq_bitmap nil 
(prog (bitmap                           % the current bitmap
             status)                    % returned status
  (cleararg 1) 
  (cleararg 2) (syscall 

  'gpr_$inq_bitmap 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq bitmap (getarg 1 '(integer32) '(fixn))) 
      (return bitmap) nil)) (t 
   
   (return (apolloerror status))))))


% ----------------------------------------------------------------------
% Allocates a bitmap in main memory and returns a bitmap descriptor
%-----------------------------------------------------------------------

(de gpr_allocate_bitmap 
                (size                   % sizes in bits for x and y
                hi_plane                % highest plane number, currently is 0
		attr)                   % descriptor of the attribute block will be
                                % used 
(prog (bitmap                           % the bitmap descriptor returned
             bitmapsize 
             status)                    % returned status

  (setq bitmapsize (gtwrds 1))          % Allocate space for Lisp Arg
  (puthalfword bitmapsize 0 (first size)) 	% Retrofit Lisp dotted pair into Pascal
  (puthalfword bitmapsize 1 (second size)) 	% record.
  (putobjpointer 1 bitmapsize) 
  (putarg 2 hi_plane '(integer16)) 
  (putarg 3 attr '(integer32)) 
  (cleararg 4) 
  (cleararg 5) (syscall 

  'gpr_$allocate_bitmap 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq bitmap (getarg 4 '(integer32) '(fixn))) 
      (return bitmap))) (t 
    (return (apolloerror status))))))


%  GPR_DEALLOCATE_BITMAP deallocates a bitmap allocated by ALLOCATE_BITMAP.

(de gpr_deallocate_bitmap 
                 (bitmap)               % the bitmap to kill

(prog (status) 
  (putarg bitmap '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$_deallocate_bitmap 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


%  GPR_INQ_BITMAP_POINTER returns a pointer to the storage/display memory
%  for the given bitmap and the number of words each scan line occupies.
%  This information can be used to directly manipulate the bits in the bitmap.

(de gpr_inq_bitmap_pointer 
                  (bitmap)              % the given bitmap

(prog (storage_ptr                      % the address of storage
             line_width                 % number of 16-bit words per line
             status)                    % returned status
  (putarg 1 bitmap '(integer32)) 
  (cleararg 2) 
  (cleararg 3) 
  (cleararg 4) (syscall 

  'gpr_$inq_bitmap_pointer 4) 
  (setq status (getarg 4 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq storage_ptr (getarg 2 '(integer32) '(fixn))) 
      (setq line_width (getarg 3 '(integer16) '(inum))) 
      (return (list storage_ptr line_width)) nil)) (t 
    
      (return (apolloerror status))))))



%  GPR_INQ_BITMAP_DIMENSIONS returns the size and number of planes
%  of the given bitmap.

(de gpr_inq_bitmap_dimensions 
                (bitmap)                % the given bitmap

(prog (size                             % sizes in bits for x and y
              hi_plane                  % highest plane number bitmap has
              status) 
  (putarg 1 bitmap '(integer32)) 
  (setq size (gtwrds 1)) 
  (putobjpointer 2 size) 
  (cleararg 3) 
  (cleararg 4) (syscall 

  'gpr_$inq_bitmap_dimensions 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq size (cons (gethalfword size 0) (gethalfword size 1))) 
      (setq hi_plane (getarg 3 '(integer16) '(inum))) 
      (return (cons size hi_plane)) nil)) (t 
    (return (apolloerror status))))))


%  GPR_ALLOCATE_ATTRIBUTE_BLOCK allocates an attribute block,
%  initialized to default settings.

(de gpr_allocate_attribute_block nil 

(prog (attrib                           % new attribute block
             status) 
  (cleararg 1) 
  (cleararg 2) (syscall 

  'gpr_$allocate_attribute_block 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq attrib (getarg 1 '(integer32) '(fixn))) 
      (return attrib) nil)) (t 
    (return (apolloerror status))))))


%  GPR_DEALLOCATE_ATTRIBUTE_BLOCK deallocates an attribute block allocated
%   by ALLOCATE_ATTRIBUTE_BLOCK.

(de gpr_deallocate_attribute_block 
                (attrib)                % attribute block to be killed

(prog (status) 
  (putarg 1 attrib '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$deallocate_attribute_block 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))




%  GPR_SET_ATTRIBUTE_BLOCK establishes the given attributes as the attributes
%   of the current bitmap.

(de gpr_set_attribute_block 
                  (attrib)              % new current attribute block

(prog (status) 
  (putarg 1 attrib '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$set_attribute_block 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))




%  GPR_ATTRIBUTE_BLOCK returns as the function value a descriptor of
%   the attributes of the given bitmap.

(de gpr_attribute_block 
                  (bitmap)              % the bitmap whose attributes are wanted
            


(prog (status 
             tempatrdesc) 
  (putarg 1 bitmap '(integer32)) 
  (cleararg 2) 

  (setq tempatrdesc (syscall 'gpr_$attribute_block 2)) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))

%eject;

%
% Graphics Primatives calls for Apollo PSL.  (second file).
%
% Author:      John W. Peterson
%              Computer Science Dept.
%              University of Utah
% Date:        December 1982
%
%  Based on:
%  GPR.INS.PAS, /sys/ins, dpg, 04/01/82
%  graphics primitive package definitions
%  by Apollo Comp. Co.
%
% 12-jan-82, Incorporated changes by Fuh-Meei Chen
%
% For further documentation on the GPR package see chapter 12 of the Apollo
%   System Programmer's Guide.
%

%  Set and inquire operations for bitmap attributes.



%  GPR_SET_attribute sets an attribute in the current bitmap.
%   The list of calls follows.
%
              
%
% Warning- the value for a boolean true in Pascal is -1 in lisp!

% Sets clipping window for current bitmap.  A window is a list of
% (BaseX BaseY SizeX SizeY)

(de gpr_set_clip_window (window) 
(prog (status windowptr) 
  (setq windowptr (gtwrds 2)) 
  (puthalfword windowptr 0 (first window)) 
  (puthalfword windowptr 1 (second window)) 
  (puthalfword windowptr 2 (third window)) 
  (puthalfword windowptr 3 (fourth window)) 
  (putobjpointer 1 windowptr) 
  (cleararg 2) (syscall 

  'gpr_$set_clip_window 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de gpr_set_clipping_active 
                (active)                % false --> off, true --> on

(prog (status) 
  (putarg 1 active '(integer16)) 
  (cleararg 2) (syscall 

  'gpr_$set_clipping_active 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 
  (cond ((eq status 0) (return status)) (t (return (apolloerror status))))))


(de gpr_set_text_font 
                  (font_id)             % new text font
(prog (status) 
  (putarg 1 font_id '(integer16)) 
  (cleararg 2) (syscall 

  'gpr_$set_text_font 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de gpr_set_coordinate_origin 
                  (origin)              % new coordinate origin

(prog (status cursororg) 
  (setq cursororg (gtwrds 1))           % Allocate space for Lisp Arg
  (puthalfword cursororg 0 (first origin)) 	%Retrofit Lisp dotted pair into Pascal
  (puthalfword cursororg 1 (second origin)) 	% record.
  (putobjpointer 1 cursororg) 
  (cleararg 2) (syscall 

  'gpr_$set_coordinate_origin 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de gpr_set_plane_mask 
                  (mask)                % new plane mask
(prog (status) 
  (putarg 1 mask '(integer16)) 
  (cleararg 2) (syscall 

  'gpr_$set_plane_mask 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



(de gpr_set_draw_value 
                  (val)                 % new line-drawing value


(prog (status) 
  (putarg 1 val '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$set_draw_value 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de gpr_set_text_value 
                  (val)                 % new text value

(prog (status) 
  (putarg 1 val '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$set_text_value 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de gpr_set_text_background_value 
                  (val)                 % new text background value

(prog (status) 
  (putarg 1 val '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$set_text_background_value 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de gpr_set_raster_op 
                (plane                  % plane for new raster op
                op)                     % new raster op

(prog (status) 
  (putarg 1 plane '(integer16)) 
  (putarg 2 op '(integer16)) 
  (cleararg 3) (syscall 

  'gpr_$set_raster_op 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))




%  GPR_INQ_attributes returns the current settings of a group
%  of attributes for the current bitmap.

(de gpr_inq_constraints nil 

(prog (window                           % clipping window
                 active                 % whether clipping is active
                 mask                   % plane mask
                 status)                % returned status

  (setq window (gtwrds 1)) 
  (putobjpointer 1 window) 
  (cleararg 2) 
  (cleararg 3) 
  (cleararg 4) (syscall 

  'gpr_$inq_constraints 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq window (cons (gethalfword window 0) (gethalfword window 1))) 
      (setq active (getarg 2 '(integer16) '(inum))) 
      (setq mask (getarg 3 '(integer16) '(inum))) 
      (return (list window active mask)) nil)) (t 
    (return (apolloerror status))))))



(de gpr_inq_text nil 

(prog (font_id                          % font id
              path                      % text path
	      status)                   % returned status

  (cleararg 1) 
  (cleararg 2) 
  (cleararg 3) (syscall 

  'gpr_$inq_text 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq font_id (getarg 1 '(integer16) '(inum))) 
      (setq path (getarg 2 '(integer16) '(inum))) 
      (return (cons font_id path)) nil)) (t 
    (return (apolloerror status))))))


(de gpr_inq_coordinate_origin nil 
(prog (origin                           % the coordinate origin
              status)                   % returned status

  (setq origin (gtwrds 1)) 
  (putobjpointer 1 origin) 
  (cleararg 2) (syscall 

  'gpr_$inq_coordinate_origin 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq origin (cons (gethalfword origin 0) (gethalfword origin 1))) 
      (return origin) nil)) (t 
    (return (apolloerror status))))))



(de gpr_inq_draw_value nil 
(prog (val                              % line-drawing value
              status)                   % returned status
  (cleararg 1) 
  (cleararg 2) (syscall 

  'gpr_$inq_draw_value 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq val (getarg 1 '(integer32) '(inum))) 
      (return val) nil)) (t 
    (return (apolloerror status))))))


(de gpr_inq_text_values nil 
(prog 
    (tval                               % text value
    bval                                % text background value
    status)                             % returned status

  (cleararg 1) 
  (cleararg 2) 
  (cleararg 3) (syscall 

  'gpr_$inq_text_values 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq tval (getarg 1 '(integer32) '(inum))) 
      (setq bval (getarg 2 '(integer32) '(inum))) 
      (return (cons tval bval)) nil)) (t 
    (return (apolloerror status))))))


%-------------------------------------------------------------------
% Returns the raster operation in use for the current bitmap
%-------------------------------------------------------------------
(de gpr_inq_raster_ops nil 
(prog (status raster_op) 
      (cleararg 1) 
      (cleararg 2) 
      (syscall 'gpr_$inq_raster_ops 2) 
      (setq status (getarg 2 '(integer32) '(fixn))) 
      (cond ((equal status 0) 
         (progn (apolloresetstatus) 
                 (setq raster_op (getarg 1 '(integer16) '(inum))) 
                 (return raster_op))) (t 
         (return (apolloerror status))))))


%eject;

%  Drawing operations.



%  GPR_MOVE sets the CP to the given position.

(de gpr_move 
                  (x y)                 % where to go
(prog (status)                          % returned status
  (putarg 1 x '(integer16)) 
  (putarg 2 y '(integer16)) 
  (cleararg 3) (syscall 

  'gpr_$move 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


%  GPR_INQ_CP returns the current position.

(de gpr_inq_cp nil 

(prog (x y                              % x and y of current position
               status)                  % returned status
  (cleararg 1) 
  (cleararg 2) 
  (cleararg 3) (syscall 

  'gpr_$inq_cp 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq x (getarg 1 '(integer16) '(inum))) 
      (setq y (getarg 2 '(integer16) '(inum))) 
      (return (cons x y)) nil)) (t 
    (return (apolloerror status))))))




%  GPR_LINE draws a line from the CP to the given position
%   and sets the CP to the given position.

(de gpr_line (x y) 

(prog (status)                          % returned status

  (putarg 1 x '(integer16)) 
  (putarg 2 y '(integer16)) 
  (cleararg 3) (syscall 

  'gpr_$line 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))




%  GPR_POLYLINE does a series of LINEs, starting from the CP.

%SysCall PROCEDURE   gpr_polyline (
%                IN  x, y: UNIV gpr_coordinate_array_t;% the list of endpoints
%                IN  npoints: integer;       % how many
%                OUT status: status_t       % returned status
%            ); EXTERN;
% This has also been left out, due to ignorance of how the hell a "Univ"
%  array is supposed to work.  Perhaps best just to take a list and use
%  GPR_LINE to do the trick??

%  GPR_MULTILINE does a series of alternate MOVEs and LINEs.

%SysCall PROCEDURE   gpr_multiline (
%                IN  x, y: UNIV gpr_coordinate_array_t;% the list of endpoints
%                IN  npoints: integer;       % how many
%                OUT status: status_t       % returned status
%            ); EXTERN;
% Again, same story as above (except use move/draw instead of just draw).
%
%eject;
%  Text operations.



%  GPR_LOAD_FONT_FILE loads a font contained in a file into an appro-
%  priate area (based on the current display mode and configuration).

(de gpr_load_font_file 
                  (pn                   % pathname of file
                  pnlen)                % pathname length

(prog (font_id                          % returned font id
                status)                 % returned status
  (putarg 1 pn '(string))               % array of 4 chars
  (putarg 2 pnlen '(integer16)) (syscall 

  'gpr_$load_font_file 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq font_id (getarg 3 '(integer16) '(inum))) 
      (return font_id) nil)) (t 
    (return (apolloerror status))))))


%  GPR_UNLOAD_FONT_FILE unloads a font that has been loaded by
%  LOAD_FONT_FILE.

(de gpr_unload_font_file 
                  (font_id)             % font id, from load_font_file

(prog (status)                          % returned status
  (putarg 1 font_id '(integer16)) 
  (cleararg 2) (syscall 

  'gpr_$unload_font_file 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



%  GPR_TEXT writes text to a bitmap from the current position ("CP").

(de gpr_text 
                 (str                   % the string to write
                 strl)                  % how long it is

(prog (status)                          % returned status
  (putarg 1 str '(string))              % array of 4 chars
  (putarg 2 strl '(integer16)) 
  (cleararg 3) (syscall 

  'gpr_$text 3) 

  (setq status (getarg 3 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))





%  GPR_INQ_TEXT_EXTENT returns the x- and y-offsets the given string
%  would span if written with TEXT.

(de gpr_inq_text_extent 
                  (str                  % the string to be inquired about
                  strl)                 % how long it is


(prog (status                           % returned status
   size)                                % how big it would be

  (putarg 1 str '(string))              % array of 4 chars
  (putarg 2 strl '(integer16)) 
  (setq size (gtwrds 1)) 
  (putobjpointer 3 size) 
  (cleararg 4) (syscall 

  'gpr_$inq_text_extent 4) 

  (setq status (getarg 4 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq size (cons (gethalfword size 0) (gethalfword size 1))) 
      (return size) nil)) (t 
    (return (apolloerror status))))))




%  GPR_INQ_TEXT_OFFSET returns the x- and y-offsets that must be added to the
%   coordinates of the desired upper left pixel of the string to give the pixel
%   from which the TEXT call should be made, and the x_offset of the pixel
%   which would be the updated CP.

(de gpr_inq_text_offset 
                  (str                  % the string to be inquired about
                strl)                   % how long it is

(prog 
         (start                         % where it would start relative to upper left pixel of string
                x_end                   % where next write would start relative to same
                status)                 % returned status

  (putarg 1 str '(string))              % array of 4 chars
  (putarg 2 strl '(integer16)) 
  (setq start (gtwrds 1)) 
  (putobjpointer 3 start) 
  (cleararg 4) 
  (cleararg 5) (syscall 

  'gpr_$inq_text_extent 5) 

  (setq status (getarg 5 '(integer32) '(fixn))) 
  (cond ((eq status 0) 
    (progn 
      (apolloresetstatus) 
      (setq start (cons (gethalfword start 0) (gethalfword start 1))) 
      (setq x_end (getarg 4 '(integer16) '(inum))) 
      (return (cons start x_end)) nil)) (t 
    (return (apolloerror status))))))



%eject;
%  Data transfer operations.



%  GPR_CLEAR clears the current bitmap to a given pixel value.

(de gpr_clear 
                  (val)                 % value to set whole bitmap to
(prog (status)                          % returned status

  (putarg 1 val '(integer32)) 
  (cleararg 2) (syscall 

  'gpr_$clear 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))




%  GPR_READ_PIXELS reads the pixels from the given window of the current
%   bitmap and stores them in a pixel array.

%SysCall PROCEDURE   gpr_read_pixels (
%                IN  src_w: gpr_window_t;   % the source window
%                OUT pix: UNIV gpr_pixel_array_t;% the pixel array
%                OUT status: status_t       % returned status
%            ); EXTERN;
%  Screw it, it only works in borrow mode


%  GPR_WRITE_PIXELS writes the pixels from a pixel array into the given
%   window of the current bitmap.

%SysCall PROCEDURE   gpr_write_pixels (
%                IN  pix: UNIV gpr_pixel_array_t; % the pixel array
%                IN  dst_w: gpr_window_t;   % the destination window
%                OUT status: status_t       % returned status
%            ); EXTERN;
%  Screw it, it only works in borrow mode.
%eject;
%  BLT operations.


% --------------------------------------------------------------------
% Perform a pixel block transfer from any bitmap to the current bitmap, 
% which is in either display or main memory
%---------------------------------------------------------------------
(de gpr_pixel_blt 
                  (src_b                % the source bitmap
                  src_w                 % the source window
                  dst_o)                % the destination origin (in current bitmap)

(prog (status                           % returned status
   window dst) 
  (putarg 1 src_b '(integer32)) 
  (setq window (gtwrds 2))              % Allocate space for Lisp Arg, two words long
  (puthalfword window 0 (first src_w))  % Retrofit Lisp dotted pair into Pascal
  (puthalfword window 1 (second src_w)) % record.
  (puthalfword window 2 (third src_w)) 
  (puthalfword window 3 (fourth src_w)) 
  (putobjpointer 2 window) 
  (setq dst (gtwrds 1)) 
  (puthalfword dst 0 (first dst_o))     % Retrofit Lisp dotted pair into Pascal
  (puthalfword dst 1 (second dst_o))    % record.
  (putobjpointer 3 dst) 
  (cleararg 4) (syscall 
  'gpr_$pixel_blt 4) 
  (setq status (getarg 4 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


% fixes for folks at amoco added..
(de gpr_bit_blt 
                  (src_b                % the source bitmap
                  src_w                 % the source window
                  src_p                 % the source plane
                  dst_o                 % the destination origin
                  dst_p)                % the destination plane
                  
(prog (window dst 
  status)                               % returned status

  (putarg 1 src_b '(integer32)) 
  (setq window (gtwrds 2))              % Allocate space for Lisp Arg
  (puthalfword window 0 (first src_w))  % Retrofit Lisp dotted pair into Pascal
  (puthalfword window 1 (second src_w)) % record.
  (puthalfword window 2 (third src_w)) 
  (puthalfword window 3 (fourth src_w)) 
  (putobjpointer 2 window) 

  (putarg 3 src_p '(integer16)) 
  (setq dst (gtwrds 1)) 
  (puthalfword dst 0 (first dst_o))     % Retrofit Lisp dotted pair into Pascal
  (puthalfword dst 1 (second dst_o))    % record.
  (putobjpointer 4 dst) 
  (putarg 5 dst_p '(integer16)) 

  (setq status (getarg 6 '(integer32) '(fixn))) (syscall 
  'gpr_$bit_blt 6) 
  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



%eject;

(de gpr_rectangle (rect) 
(prog (status windowspecs)              % basex,basey, sizex,sizey
  (setq windowspecs (gtwrds 2)) 
  (puthalfword windowspecs 0 (first rect)) 
  (puthalfword windowspecs 1 (second rect)) 
  (puthalfword windowspecs 2 (third rect)) 
  (puthalfword windowspecs 3 (fourth rect)) 
  (putobjpointer 1 windowspecs) 
  (cleararg 2) 

  (syscall 'gpr_$rectangle 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


(de gpr_set_fill_value (index) 
(prog (status) 
  (putarg 1 index '(integer32)) 
  (cleararg 2) 
  (syscall 'gpr_$set_fill_value 2) 
  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


             
(de gpr_color_init nil 
(prog (init_id fonts_id graphics_channel) 
                        % create a pad on top of standard transcript pad
   (setq graphics_channel 
      (pad_create "" 0 1 2 0 75))       % name of edited file, blank for tempory file
                  % transcript style pad
                  % the window on which this pad is open, standoutput
                  % position this pad at the top of the current
                            % window 
                  % Pane size is relative
                  % pane size is 75%
 
                      % sets a scale factor for cursor operations
      (pad_set_scale graphics_channel 1 1) 	% the stream id on which the pad is open
                     % the scale factor for the x-coordinate  
                     % the scale factor for the y-coordinate

                      % creates a frame in a pad
      (pad_create_frame graphics_channel 800 1000) 	% the stream id on which
                                           % the pad is open
                        % the width of the frame
                        % the height of the frame
   
                      % initialize the graphics primitives package
      (setq init_id 
         (gpr_init 1                    % frame mode
                   graphics_channel     % stream id of the pad
                   (list 800 1000) 7))  % size of the initial bitmap
                   % currently is  (8 planes; 255 colors)
                       
         % establishes a bitmap as the current bit map for subsequent operation
      (gpr_set_bitmap init_id) 
         
         % loads a font file for the graphics to use

      (setq fonts_id (pad_load_font graphics_channel "/sys/dm/fonts/std")) 

      (setq fonts_id (gpr_load_font_file "/sys/dm/fonts/std" 17)) 

         % establishes a new font for subsequent text operation
      (gpr_set_text_font fonts_id) 
      (return (cons graphics_channel fonts_id))))


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Direct mode stuff
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%   Acquires use of the display in direct graphics mode and tells
%   whether the window was obscured.

(de gpr_acquire_display nil 
(prog (status obscured) 
  (cleararg 1) 

  (setq obscured (syscall 'gpr_$acquire_display 1)) 

  (setq status (getarg 1 '(integer32) '(fixn))) 

  (cond ((eq status 0) (progn (apolloresetstatus) (return obscured))) (t 
    (return (apolloerror status))))))



%   Establishes whether the display manager will automatically save and
%   refresh the contents of an acquired display window in direct graphics
%   mode.
(de gpr_set_auto_refresh (auto_refresh_flag) 
(prog (status) 
  (cond ((inump auto_refresh_flag) (putarg 1 auto_refresh_flag '(integer16))) (t 
    
      (progn 
        (cond (auto_refresh_flag (putarg 1 -1 '(integer16))) (t 
           (putarg 1 0 '(integer16))))))) 
      
  (cleararg 2) (syscall 

  'gpr_$set_auto_refresh 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


%   Force a release of an acquired display in direct graphics mode
%   and reports how many acquire operations were pending.

(de gpr_force_release nil 
(prog (acquirecount status) 
  (cleararg 1) 
  (cleararg 2) (syscall 

  'gpr_$force_release 2) 

  (setq acquirecount (getarg 1 '(integer16) '(inum))) 
  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) (progn (apolloresetstatus) (return acquirecount))) (t 
    (return (apolloerror status))))))



%
% Set action for obscured direct windows.

(de gpr_set_obscured_opt (if_obscured) 
(prog (status) 
  (putarg 1 if_obscured '(integer16)) 
  (cleararg 2) (syscall 

  'gpr_$set_obscured_opt 2) 

  (setq status (getarg 2 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



(de gpr_start_pgon (x y) 
(prog (status) 
    (putarg 1 x '(integer16)) 
    (putarg 2 y '(integer16)) 
    (cleararg 3) (syscall 

    'gpr_$start_pgon 3) 

    (setq status (getarg 3 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



(fluid '(gpr_xlist gpr_ylist))
(setq gpr_xlist (gtbps 255))
(setq gpr_ylist (gtbps 255))

%
% Expects a list ( (x1 y1) (x2 y2) ... )
(de gpr_pgon_polyline (pointlist) 
(prog (status point index) 

  (setq index 0) 
  (foreach point in pointlist do (progn 
    (puthalfword gpr_xlist index (first point)) 
    (puthalfword gpr_ylist index (second point)) 
    (setq index (plus index 1)) nil)) 
  

  (putobjpointer 1 gpr_xlist) 
  (putobjpointer 2 gpr_ylist) 
  (putarg 3 (length pointlist) '(integer16)) 
  (cleararg 4) (syscall 
              
  'gpr_$pgon_polyline 4) 
  (setq status (getarg 4 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))



(de gpr_close_fill_pgon nil 
(prog (status) 
  (cleararg 1) (syscall 
  'gpr_$close_fill_pgon 1) 
  (setq status (getarg 1 '(integer32) '(fixn))) 

  (cond ((eq status 0) 
   (progn (apolloresetstatus) (return status))) (t 
    (return (apolloerror status))))))


